import io

from ... import expression_walker as ew
from ...expression_walker import ReplaceExpressionWalker
from ...expressions import Constant, ExpressionBlock, Symbol
from ...logic import (Conjunction, ExistentialPredicate, FunctionApplication,
                      Implication)
from ...logic.transformations import CollapseConjunctions
from ..aggregation import DatalogWithAggregationMixin
from ..expressions import TranslateToLogic
from ..ontologies_parser import OntologyParser, RightImplication
from ..ontologies_rewriter import OntologyRewriter

S_ = Symbol
C_ = Constant
EP_ = ExistentialPredicate
EB_ = ExpressionBlock
FA_ = FunctionApplication
I_ = Implication
RI_ = RightImplication


class DatalogTranslator(
    TranslateToLogic, ew.IdentityWalker, DatalogWithAggregationMixin
):
    pass

def  _categorize_constraints(formulas):
    parsed_constraints = {}
    for sigma in formulas:
        sigma_functor = sigma.consequent.functor.name
        if sigma_functor in parsed_constraints:
            cons_set = parsed_constraints[sigma_functor]
            if sigma not in cons_set:
                cons_set.add(sigma)
                parsed_constraints[sigma_functor] = cons_set
        else:
            parsed_constraints[sigma_functor] = set([sigma])

    return parsed_constraints

def test_normal_rewriting_step():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")
    a = S_("a")
    b = S_("b")
    db = C_("db")

    sigma = RI_(project(x) & inArea(x, y), hasCollaborator(z, y, x))
    q = I_(p(b), hasCollaborator(a, db, b))

    qB = EB_((q,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    sigmaB = _categorize_constraints([sigma])

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 2
    imp1 = rewrite.pop()
    imp2 = rewrite.pop()
    assert imp1 == q or imp2 == q
    q2 = I_(p(b), project(b) & inArea(b, db))
    assert imp1 == q2 or imp2 == q2


def test_more_than_one_free_variable():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    w = S_("w")
    x = S_("x")
    y = S_("y")
    z = S_("z")
    a = S_("a")
    b = S_("b")
    c = S_("c")
    db = C_("db")

    sigma = RI_(project(x) & inArea(x, y), hasCollaborator(w, z, y, x))
    q = I_(p(b), hasCollaborator(c, a, db, b))

    qB = EB_((q,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    sigmaB = _categorize_constraints([sigma])

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 2
    imp1 = rewrite.pop()
    imp2 = rewrite.pop()
    assert imp1 == q or imp2 == q
    q2 = I_(p(b), project(b) & inArea(b, db))
    assert imp1 == q2 or imp2 == q2


def test_unsound_rewriting_step_constant():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")
    b = S_("b")
    db = C_("db")
    c = C_("c")

    sigma = RI_(project(x) & inArea(x, y), hasCollaborator(z, y, x))
    q = I_(p(b), hasCollaborator(c, db, b))

    qB = EB_((q,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    sigmaB = _categorize_constraints([sigma])

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 1
    imp = rewrite.pop()
    assert imp == q


def test_unsound_rewriting_step_shared():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")
    b = S_("b")
    db = C_("db")

    sigma = RI_(project(x) & inArea(x, y), hasCollaborator(z, y, x))
    q = I_(p(b), hasCollaborator(b, db, b))

    qB = EB_((q,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    sigmaB = _categorize_constraints([sigma])

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 1
    imp = rewrite.pop()
    assert imp == q


def test_outside_variable():
    s = S_("s")
    r = S_("r")
    t = S_("t")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")

    a = S_("a")
    b = S_("b")
    c = S_("c")
    e = S_("e")

    sigma = RI_(s(x) & r(x, y), t(x, y, z))
    q2 = I_(p(a), s(c) & t(a, b, c) & t(a, e, c))

    qB = EB_((q2,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    sigmaB = _categorize_constraints([sigma])

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 1


def test_example_4_3():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    collaborator = S_("collaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")
    a = S_("a")
    b = S_("b")
    c = S_("c")

    sigma1 = RI_(project(x) & inArea(x, y), hasCollaborator(z, y, x))
    sigma2 = RI_(hasCollaborator(x, y, z), collaborator(x))

    q = I_(p(b, c), hasCollaborator(a, b, c) & collaborator(a))

    qB = EB_((q,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    sigmaB = _categorize_constraints([sigma1, sigma2])

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 4


def test_infinite_walker():

    subClassOf = Symbol(str("subClassOf"))
    rest = Symbol("rest")
    reg = Symbol("reg")

    x1 = Symbol("x1")
    y1 = Symbol("y1")
    x = Symbol("x")
    y = Symbol("y")

    sigma_ant = Conjunction((subClassOf(x1, y1), rest(x1, y1), rest(y1, y1)))
    S = subClassOf(x1, y1)
    q_ant = Conjunction((subClassOf(x1, y1), rest(y, x1), rest(x, y1), reg(x)))

    replace = dict({S: sigma_ant})
    rsw = ReplaceExpressionWalker(replace)
    sigma_rep = rsw.walk(q_ant)
    sigma_rep = CollapseConjunctions().walk(sigma_rep)

    expected = Conjunction((sigma_ant, rest(y, x1), rest(x, y1), reg(x)))
    expected = CollapseConjunctions().walk(expected)

    assert sigma_rep == expected


def test_distinguished_variables():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    a = S_("a")
    b = S_("b")

    sigma = RI_(project(x), hasCollaborator(y, x))
    q = I_(p(a), hasCollaborator(a, b) & inArea(b))

    qB = EB_((q,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    sigmaB = _categorize_constraints([sigma])

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 1
    imp1 = rewrite.pop()
    assert imp1 == qB.formulas[0]

def test_empty_rewrite():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="AdministrativeStaff">
            <rdfs:label>administrative staff worker</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Employee"/>
        </owl:Class>

        <owl:Class rdf:ID="Article">
            <rdfs:label>article</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Publication"/>
        </owl:Class>

        <owl:Class rdf:ID="AssistantProfessor">
            <rdfs:label>assistant professor</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Professor"/>
        </owl:Class>

        <owl:Class rdf:ID="AssociateProfessor">
            <rdfs:label>associate professor</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Professor"/>
        </owl:Class>

        <owl:Class rdf:ID="Book">
            <rdfs:label>book</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Publication"/>
        </owl:Class>
    </rdf:RDF>'''

    book = Symbol('book')
    x = Symbol('x')
    p = Symbol('p')

    onto = OntologyParser(io.StringIO(owl))
    constraints, _, _ = onto.parse_ontology()

    q = I_(p(x), book(x))

    qB = EB_((q,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    orw = OntologyRewriter(qB, constraints)
    rewrited = orw.Xrewrite()

    assert len(rewrited) == 1
    imp = rewrited.pop()
    assert imp == q


def test_ontology_parsed_rewrite():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    headof = Symbol('headOf')
    chair = Symbol('Chair')
    x = Symbol('x')
    y = Symbol('y')
    p = Symbol('p')

    onto = OntologyParser(io.StringIO(owl))
    constraints, _, _ = onto.parse_ontology()

    q = I_(p(x), headof(x, y))

    qB = EB_((q,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)

    orw = OntologyRewriter(qB, constraints)
    rewrited = orw.Xrewrite()

    rewrited = list(rewrited)

    assert len(rewrited) == 3
    assert q in rewrited
    index_no_q = [i for i, e in enumerate(rewrited) if e != q]
    assert len(index_no_q) == 2
    index_1 = index_no_q[0]
    index_2 = index_no_q[1]

    if rewrited[index_1].antecedent == chair(x):
        assert len(rewrited[index_2].antecedent.args) == 2
        assert rewrited[index_2].antecedent.args[0] == rewrited[index_2].consequent.args[0]
    elif rewrited[index_2].antecedent == chair(x):
        assert len(rewrited[index_1].antecedent.args) == 2
        assert rewrited[index_1].antecedent.args[0] == rewrited[index_1].consequent.args[0]
    else:
        assert False



from ...config import config

try:
    if config["RAS"].getboolean("synchronous", False):
        import dask

        dask.config.set(scheduler="single-threaded")

    from dask_sql import Context
    from dask_sql.mappings import sql_to_python_type
    from dask.distributed import Client
    import dask.dataframe as dd
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "Unable to use the dask backend because dask"
        " has not been installed. Make sure to install with"
        " `pip install neurolang[dask]` to enable the dask backend."
    ) from e

import ast
import inspect
import logging
import threading
import time
import types
from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Type, Union

import numpy as np
from neurolang.type_system import (
    Unknown,
    get_args,
    infer_type_builtins,
    typing_callable_from_annotated_function,
)
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import functions

import pandas as pd
from pandas.api.types import pandas_dtype


LOG = logging.getLogger(__name__)


def timeit(func):
    """
    This decorator logs the execution time for the decorated function.
    Log times will not be acurate for Dask operations if using asynchronous
    scheduler.
    """

    def wrapper(*args, **kwargs):
        start = time.perf_counter()
        result = func(*args, **kwargs)
        elapsed = time.perf_counter() - start
        LOG.debug("======================================")
        LOG.debug(f"{func.__name__} took {elapsed:2.4f} s")
        LOG.debug("======================================")
        return result

    return wrapper


def _id_generator():
    lock = threading.RLock()
    i = 0
    while True:
        with lock:
            fresh = f"{i:08}"
            i += 1
        yield fresh


class DaskContextManager(ABC):
    """
    Singleton class to manage Dask-related objects, mainly
    Dask's Client and Dask-SQL's Context.
    """

    _context = None
    _client = None
    _id_generator_ = _id_generator()

    @abstractmethod
    def _do_not_instantiate_singleton_class(self):
        pass

    @classmethod
    def _create_client(cls):
        if cls._client is None:
            cls._client = Client(processes=False)

    @classmethod
    def get_context(cls, new=False):
        if cls._context is None or new:
            if not config["RAS"].getboolean("synchronous", False):
                cls._create_client()
            cls._context = Context()
            # We register an aggregate function called len which applies to string columns
            # Used for example in `test_probabilistic_frontend:test_postprob_conjunct_with_wlq_result`
            cls._context.register_aggregation(
                len, "len", [("x", pd.StringDtype())], np.int32
            )
            # We also register a sum which applies to objects (i.e `Symbol` or sets)
            # since by default sum applies only to numbers in SQL and Calcite will
            # try to cast objects to float before applying the default sum op.
            cls._context.register_aggregation(
                sum, "sum", [("x", np.object_)], np.object_
            )
        return cls._context

    @classmethod
    def sql(cls, query):
        compiled_query = cls.compile_query(query)
        LOG.info(f"Executing SQL query :\n{compiled_query}")
        return cls.get_context().sql(compiled_query)

    @staticmethod
    def compile_query(query):
        return str(
            query.compile(
                dialect=postgresql.dialect(),
                compile_kwargs={"literal_binds": True},
            )
        )

    @classmethod
    def register_function(cls, f_, fname, params, return_type, wrapped):
        func_to_register = f_
        if wrapped:
            func_to_register = cls.wrap_function_with_dataframe(
                f_, params, return_type
            )
        cls.get_context().register_function(
            func_to_register, fname, params, return_type
        )

    @classmethod
    def register_aggregation(cls, f_, fname, params, return_type):
        # FIXME: We should preferably try to use GroupBy-aggregations
        # instead of GroupBy-apply when doing aggregations. I.e.
        # create a dd.Aggregation from the given function and register it
        # on the context. But this doesnt work in all cases, since dask
        # applies GroupBy-aggregations first on each chunk, then again to
        # the results of all the chunk aggregations.
        # So transformative aggregation will not work properly, for
        # instance sum(x) - 1 will result in sum(x) - 2 in the end.
        # agg = dd.Aggregation(
        #     fname, lambda chunk: chunk.agg(f_), lambda total: total.agg(f_)
        # )

        func_to_register = f_
        if len(params) > 1:
            func_to_register = cls.wrap_function_with_param_names(f_, params)
        cls.get_context().register_aggregation(
            func_to_register, fname, params, return_type
        )

    @staticmethod
    def wrap_function_with_param_names(f_, params):
        try:
            pnames = [name for (name, _) in params]
            named_tuple_type = namedtuple("LambdaTuple", pnames)
        except ValueError:
            # Invalid column names, just use a tuple instead.
            named_tuple_type = None

        def wrapped_custom_function(*values):
            if named_tuple_type:
                return f_(named_tuple_type(*values))
            else:
                return f_(tuple(values))

        return wrapped_custom_function

    @staticmethod
    def wrap_function_with_dataframe(f_, params, return_type):
        """
        The way a function is called in dask_sql is by calling it with a list
        of its parameters (dask series,see dask_sql.physical.rex.core.call.py)
        Also, dask_sql is changing the names of the columns internally.
        What we want to do is transform these Series into a dataframe with the
        expected column names (params) and call apply on it.

        This is what wrapped_custom_function does.

        Concatenating Series into a Dask DataFrame is complicated because of
        unknown partitions, so we turn the first Series into a DataFrame and
        then assign the other Series as columns.
        """
        pnames = [name for (name, _) in params]

        def wrapped_custom_function(*values):
            s0 = values[0]
            if not isinstance(s0, dd.Series):
                # Sometimes the Calcite optimizer will push a constant into
                # the params of a call so we need to turn it into a Series.
                s0 = dd.from_pandas(
                    pd.Series([s0], name=pnames[0]), npartitions=1, sort=False
                )
            else:
                s0.name = pnames[0]
            ddf = s0.to_frame()
            for name, col in zip(pnames[1:], values[1:]):
                ddf[name] = col
            return ddf.apply(f_, axis=1, meta=(None, return_type))

        return wrapped_custom_function


def convert_types_to_pandas_dtype(
    types: pd.Series, default_type: np.dtype = np.float64
) -> pd.Series:
    """
    Convert a series of native python types and typing type annotations
    into a series of pandas/numpy dtypes.

    Parameters
    ----------
    types : pd.Series
        the types
    default_type : np.dtype, optional
        the default type for when a type is Unknown, by default np.float64

    Returns
    -------
    pd.Series
        the converted types
    """
    nptypes = []
    for t in types:
        nptypes.append(convert_type_to_pandas_dtype(t, default_type))
    return pd.Series(nptypes, index=types.index)


def convert_type_to_pandas_dtype(
    type_: Union[type, Type], default_type: np.dtype = np.float64
) -> np.dtype:
    """
    Convert a native python type or typing type annotation to a numpy dtype.

    Parameters
    ----------
    type_ : Union[type, Type]
        the type to convert
    default_type : np.dtype, optional
        default dtype used for when type is Unknown, by default np.float64

    Returns
    -------
    np.dtype
        the converted type
    """
    if type_ is Unknown:
        nptype = default_type
    elif isinstance(type_, type) and issubclass(type_, str):
        nptype = pd.StringDtype()
    else:
        try:
            nptype = pandas_dtype(type_)
        except TypeError:
            # assume it's an object type
            nptype = np.object_
    return nptype


def try_to_infer_type_of_operation(
    operation, column_types, default_type=np.float64
):
    """
    Tries to infer the return type for an operation passed to aggregate
    or extended_projection methods.
    In order to work with dask-sql, the return type should be a pandas
    or numpy type.

    Parameters
    ----------
    operation : Union[Callable, str]
        The operation to infer the type for
    column_types : pd.Series
        The dtypes series mapping the dtype for each column.
        Used if operation references a known column.
    default_type : Type, optional
        The return value if type cannot be infered, by default np.float64

    Returns
    -------
    Type
        An infered return type for the operation.
    """
    try:
        # 1. First we try to guess the return type of the operation
        if isinstance(operation, (types.FunctionType, types.MethodType)):
            # operation is a custom function
            rtype = typing_callable_from_annotated_function(operation)
            rtype = get_args(rtype)[1]
        elif isinstance(operation, types.BuiltinFunctionType):
            # operation is something like 'sum'
            rtype = infer_type_builtins(operation)
            rtype = get_args(rtype)[1]
        else:
            if isinstance(operation, str):
                default_type = str
                # check if it's one of SQLAlchemy's known functions, like count
                if hasattr(functions, operation):
                    rtype = getattr(functions, operation).type
                    if inspect.isclass(rtype):
                        rtype = rtype()
                    rtype = sql_to_python_type(rtype.compile())
                else:
                    # otherwise operation is probably a str or
                    # RelationalAlgebraStringExpression representing a column
                    # literal, like 'col_a + 1', or a constant like '0'.
                    # We try to parse the expression to get type of variable or
                    # constant.
                    rtype = type_of_expression(
                        ast.parse(operation, mode="eval").body, column_types
                    )
            else:
                rtype = type(operation)
    except (ValueError, TypeError, NotImplementedError, SyntaxError):
        LOG.warning(
            f"Unable to infer type of operation {operation}"
            f", assuming default {default_type} type instead."
        )
        rtype = default_type
    return rtype


def type_of_expression(node, column_types):
    if isinstance(node, ast.Num):  # <number>
        return type(node.n)
    elif isinstance(node, ast.Constant):  # Constant
        return type(node.value)
    elif isinstance(node, ast.Name):
        if node.id in column_types.index:  # Col name
            return column_types[node.id]
        else:
            return str
    elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return type_of_expression(node.left, column_types)
    elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return type_of_expression(node.operand, column_types)
    else:
        return Unknown



import logging
from functools import reduce
from itertools import chain, combinations
from typing import AbstractSet

import numpy as np

from .. import relational_algebra_provenance as rap
from ..config import config
from ..datalog.expression_processing import (
    UnifyVariableEqualities,
    flatten_query
)
from ..datalog.translate_to_named_ra import TranslateToNamedRA
from ..exceptions import (
    NeuroLangException,
    NonLiftableException,
    NotInFONegE,
    NotRankedException,
    NotUnateException,
    UnsupportedSolverError
)
from ..expression_walker import (
    ExpressionWalker,
    PatternWalker,
    ReplaceExpressionWalker,
    add_match,
    expression_iterator
)
from ..expressions import (
    Constant,
    FunctionApplication,
    Symbol,
    TypedSymbolTableMixin
)
from ..logic import (
    FALSE,
    Conjunction,
    Disjunction,
    ExistentialPredicate,
    Implication,
    NaryLogicOperator,
    Negation,
    Quantifier,
    Union,
    UniversalPredicate
)
from ..logic.expression_processing import (
    extract_logic_atoms,
    extract_logic_free_variables,
    extract_logic_predicates
)
from ..logic.transformations import (
    ExtractConjunctiveQueryWithNegation,
    GuaranteeConjunction,
    GuaranteeDisjunction,
    MakeExistentialsImplicit,
    MakeUniversalsImplicit,
    MoveNegationsToAtoms,
    MoveQuantifiersUp,
    PushQuantifiersDown,
    RemoveExistentialOnVariables,
    RemoveTrivialOperations,
    RemoveUniversalPredicates,
    convert_to_pnf_with_dnf_matrix
)
from ..relational_algebra import (
    BinaryRelationalAlgebraOperation,
    ColumnStr,
    NamedRelationalAlgebraFrozenSet,
    NAryRelationalAlgebraOperation,
    UnaryRelationalAlgebraOperation,
    str2columnstr_constant
)
from ..relational_algebra_provenance import ProvenanceAlgebraSet
from ..utils import OrderedSet, log_performance
from .containment import is_contained
from .expression_processing import (
    is_builtin,
    lift_optimization_for_choice_predicates
)
from .probabilistic_ra_utils import (
    NonLiftable,
    ProbabilisticChoiceSet,
    ProbabilisticFactSet,
    generate_probabilistic_symbol_table_for_query,
    is_atom_a_deterministic_relation,
    is_atom_a_probabilistic_choice_relation
)
from .probabilistic_semiring_solver import (
    ProbSemiringToRelationalAlgebraSolver
)
from .query_resolution import (
    compute_projections_needed_to_reintroduce_head_terms,
    generate_provenance_query_solver,
    lift_solve_marg_query
)
from .ranking import partially_rank_query, verify_that_the_query_is_ranked
from .shattering import shatter_constants
from .transforms import (
    add_existentials_except,
    convert_rule_to_ucq,
    convert_to_cnf_existential_ucq,
    convert_to_dnf_existential_ucq,
    minimize_ucq_in_cnf,
    minimize_ucq_in_dnf,
    unify_existential_variables
)

LOG = logging.getLogger(__name__)


__all__ = [
    "dalvi_suciu_lift",
    "solve_succ_query",
    "solve_marg_query",
]


GC = GuaranteeConjunction()
MNTA = MoveNegationsToAtoms()
MQU = MoveQuantifiersUp()
PQD = PushQuantifiersDown()
RTO = RemoveTrivialOperations()
RUP = RemoveUniversalPredicates()


class ExtendedRAPToRAWalker(
    rap.IndependentDisjointProjectionsAndUnionMixin,
    rap.WeightedNaturalJoinSolverMixin,
    ProbSemiringToRelationalAlgebraSolver,
):
    pass


class ReplaceAllSymbolsButConstantSets(ExpressionWalker):
    def __init__(self, symbol_table):
        self.symbol_table = symbol_table

    @add_match(Symbol)
    def replace_symbol(self, expression):
        res = expression
        if expression in self.symbol_table:
            value = self.symbol_table[expression]
            if not isinstance(value, Constant[AbstractSet]):
                res = value
        return res


def solve_succ_query(query, cpl_program, run_relational_algebra_solver=True):
    """
    Solve a SUCC query on a CP-Logic program.

    Parameters
    ----------
    query : Implication
        SUCC query of the form `ans(x) :- P(x)`.
    cpl_program : CPLogicProgram
        CP-Logic program on which the query should be solved.
    run_relational_algebra_solver: bool
        When true the result's `relation` attribute is a
        NamedRelationalAlgebraFrozenSet,
        when false the attribute is the relational algebra expression that
        produces the such set.

    Returns
    -------
    ProvenanceAlgebraSet
        Provenance set labelled with probabilities for each tuple in the result
        set.

    """
    with log_performance(
        LOG,
        "Preparing query %s",
        init_args=(query.consequent.functor.name,),
    ):
        flat_query_body = flatten_query(query.antecedent, cpl_program)

    if flat_query_body == FALSE or (
        isinstance(flat_query_body, Conjunction)
        and any(conjunct == FALSE for conjunct in flat_query_body.formulas)
    ):
        head_var_names = tuple(
            term.name for term in query.consequent.args
            if isinstance(term, Symbol)
        )
        return ProvenanceAlgebraSet(
            Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
                ("_p_",) + head_var_names
            )),
            str2columnstr_constant("_p_"),
        )

    with log_performance(LOG, "Translation to extensional plan"):
        flat_query = Implication(query.consequent, flat_query_body)
        shattered_query, symbol_table = \
            _prepare_and_optimise_query(flat_query, cpl_program)
        ucq_shattered_query = convert_rule_to_ucq(shattered_query)
        ucq_shattered_query = MNTA.walk(
            MQU.walk(ucq_shattered_query)
        )
        ra_query = dalvi_suciu_lift(ucq_shattered_query, symbol_table)
        if not is_pure_lifted_plan(ra_query):
            LOG.info(
                "Query not liftable %s",
                shattered_query
            )
            raise NonLiftableException(
                "Query %s not liftable, algorithm can't be applied",
                query
            )
        ra_query = (
            ReplaceAllSymbolsButConstantSets(symbol_table)
            .walk(ra_query)
        )

    needed_projections = compute_projections_needed_to_reintroduce_head_terms(
        ra_query, flat_query, shattered_query
    )

    query_solver = generate_provenance_query_solver(
        symbol_table, run_relational_algebra_solver,
        needed_projections=needed_projections,
        solver_class=ExtendedRAPToRAWalker
    )

    with log_performance(LOG, "Run RAP query %s", init_args=(ra_query,)):
        prob_set_result = query_solver.walk(ra_query)

    return prob_set_result


def _prepare_and_optimise_query(flat_query, cpl_program):
    flat_query_body = convert_to_dnf_existential_ucq(flat_query.antecedent)
    flat_query_body = RTO.walk(Disjunction(tuple(
        lift_optimization_for_choice_predicates(f, cpl_program)
        for f in flat_query_body.formulas
    )))
    flat_query = Implication(flat_query.consequent, flat_query_body)
    unified_query = UnifyVariableEqualities().walk(flat_query)
    if config.get_probabilistic_solver_check_unate():
        _verify_that_the_query_has_one_quantifier(flat_query)
    symbol_table = generate_probabilistic_symbol_table_for_query(
        cpl_program, unified_query.antecedent
    )
    shattered_query_body = shatter_constants(
        convert_rule_to_ucq(unified_query),
        symbol_table
    )
    shattered_query_body = partially_rank_query(
        shattered_query_body, symbol_table
    )
    shattered_query = Implication(
        unified_query.consequent,
        shattered_query_body
    )
    _verify_that_the_query_has_no_builtins(shattered_query, cpl_program)
    if config.get_probabilistic_solver_check_unate():
        _verify_that_the_query_is_unate(shattered_query, symbol_table)
    if not verify_that_the_query_is_ranked(
        convert_rule_to_ucq(shattered_query)
    ):
        raise NotRankedException(f"Query {flat_query} is not ranked")

    return shattered_query, symbol_table


def _verify_that_the_query_has_no_builtins(shattered_query, cpl_program):
    if any(
        is_builtin(atom, known_builtins=cpl_program.builtins())
        for atom in extract_logic_atoms(shattered_query)
    ):
        raise UnsupportedSolverError(
            "Builtins not allowed for Dalvi-Suciu lifting"
        )


def _verify_that_the_query_is_unate(query, symbol_table):
    positive_relational_symbols = set()
    negative_relational_symbols = set()

    query = convert_rule_to_ucq(query)
    query = convert_to_pnf_with_dnf_matrix(query)
    for predicate in extract_logic_predicates(query):
        if isinstance(predicate, Negation):
            atom = predicate.formula
        else:
            atom = predicate
        if is_atom_a_deterministic_relation(atom, symbol_table):
            continue
        if isinstance(predicate, Negation):
            while isinstance(predicate, Negation):
                predicate = predicate.formula
            negative_relational_symbols.add(predicate.functor)
        else:
            positive_relational_symbols.add(predicate.functor)

    if not positive_relational_symbols.isdisjoint(negative_relational_symbols):
        raise NotUnateException(f"Query {query} is not unate")


def _verify_that_the_query_has_one_quantifier(query):
    query = convert_rule_to_ucq(query)
    query = convert_to_pnf_with_dnf_matrix(query)
    if not isinstance(query, Quantifier):
        return

    first_quantifier = type(query)
    query_act = query.body
    while isinstance(query_act, Quantifier):
        if not isinstance(query_act, first_quantifier):
            raise NotUnateException(
                f"Query {query} uses two types of quantifiers"
            )
        query_act = query_act.body


def solve_marg_query(rule, cpl):
    return lift_solve_marg_query(rule, cpl, solve_succ_query)


def dalvi_suciu_lift(rule, symbol_table):
    '''
    Translation from a datalog rule which allows disjunctions in the body
    to a safe plan according to [1]. Non-liftable segments are identified
    by the `NonLiftable` expression.
    [1] Dalvi, N. & Suciu, D. The dichotomy of probabilistic inference
    for unions of conjunctive queries. J. ACM 59, 1–87 (2012).
    '''
    #  rule = RemoveUniversalPredicates().walk(rule)
    rule = RTO.walk(rule)

    has_safe_plan, res = symbol_or_deterministic_plan(rule, symbol_table)
    if has_safe_plan:
        return res

    rule_cnf = convert_ucq_to_ccq(rule, transformation='CNF')
    connected_components = symbol_connected_components(rule_cnf)
    if len(connected_components) > 1:
        return components_plan(
            connected_components, rap.NaturalJoin, symbol_table,
            negative_operation=rap.Difference
        )

    rule_dnf = convert_ucq_to_ccq(rule, transformation='DNF')
    connected_components = symbol_connected_components(rule_dnf)
    if len(connected_components) > 1:
        return components_plan(
            connected_components, rap.Union, symbol_table
        )

    rule_pdnf = convert_to_pnf_with_dnf_matrix(rule_dnf)
    has_svs, plan = has_separator_variables(rule_pdnf, symbol_table)
    if has_svs:
        return plan

    has_safe_plan, plan = disjoint_project(rule_dnf, symbol_table)
    if has_safe_plan:
        return plan

    if len(rule_cnf.formulas) > 1:
        return inclusion_exclusion_conjunction(rule_cnf, symbol_table)

    return NonLiftable(rule)


def symbol_or_deterministic_plan(rule, symbol_table):
    """If the rule is a single relational symbol or a
    fully deterministic subquery, then return the safe plan.

    Parameters
    ----------
    rule : Expression
        rule to lift
    symbol_table : Mapping of Symbol to Expression
        symbol table for the relational symbols

    Returns
    -------
        Tuple[bool, Union[Expression, NoneType]]
        if the rule is liftable then the boolean will be True and
        the second element of the tuple will be the lifted plan.
    """

    has_safe_plan = False
    res = None
    if isinstance(rule, FunctionApplication):
        has_safe_plan = True
        res = TranslateToNamedRA().walk(rule)
    elif (
        all(
            is_atom_a_deterministic_relation(atom, symbol_table)
            for atom in extract_logic_atoms(rule)
        )
    ):
        free_vars = extract_logic_free_variables(rule)
        rule = MakeExistentialsImplicit().walk(rule)
        result = TranslateToNamedRA().walk(rule)
        proj_cols = tuple(Constant(ColumnStr(v.name)) for v in free_vars)
        has_safe_plan = True
        res = rap.Projection(result, proj_cols)
    return has_safe_plan, res


def convert_ucq_to_ccq(rule, transformation='CNF'):
    """Given a UCQ expression, this function transforms it into a connected
    component expression following the definition provided by Dalvi Suciu[1]
    in section 2.6 Special Query Expressions. The transformation can be
    parameterized to apply a CNF or DNF transformation, as needed.

    [1] Dalvi, N. & Suciu, D. The dichotomy of probabilistic inference
    for unions of conjunctive queries. J. ACM 59, 1–87 (2012).

    Parameters
    ----------
    rule : Definition
        UCQ expression

    Returns
    -------
    NAryLogicOperation
        Transformation of the initial expression
        in a connected component query.
    """
    rule = PQD.walk(RUP.walk(PQD.walk(MNTA.walk(rule))))
    free_vars = extract_logic_free_variables(rule)
    existential_vars = set()
    for atom in extract_logic_atoms(rule):
        existential_vars.update(set(atom.args) - set(free_vars))

    conjunctions = ExtractConjunctiveQueryWithNegation().walk(rule)
    dic_components = extract_connected_components(
        conjunctions, existential_vars
    )

    fresh_symbols_expression = (
        ReplaceExpressionWalker(dic_components)
        .walk(rule)
    )
    if transformation == 'CNF':
        fresh_symbols_expression = convert_to_cnf_existential_ucq(
            fresh_symbols_expression
        )
        minimize = minimize_ucq_in_cnf
        gcd = GuaranteeConjunction()
    elif transformation == 'DNF':
        fresh_symbols_expression = convert_to_dnf_existential_ucq(
            fresh_symbols_expression
        )
        minimize = minimize_ucq_in_dnf
        gcd = GuaranteeDisjunction()
    else:
        raise ValueError(f'Invalid transformation type: {transformation}')

    final_expression = ReplaceExpressionWalker(
        {v: k for k, v in dic_components.items()}
    ).walk(fresh_symbols_expression)

    try:
        final_expression = minimize(final_expression)
    except NotInFONegE:
        pass

    return gcd.walk(final_expression)


def extract_connected_components(list_of_conjunctions, existential_vars):
    """Given a list of conjunctions, this function is in charge of calculating
    the connected components of each one. As a result, it returns a dictionary
    with the necessary transformations so that these conjunctions are replaced
    by symbols that preserve their free variables.

    Parameters
    ----------
    list_of_conjunctions : list
        List of conjunctions for which we want to
        calculate the connected components

    existential_vars : set
        Set of variables associated with existentials
        in the expressions that compose the list of conjunctions.

    Returns
    -------
    dict
        Dictionary of transformations where the keys are the expressions that
        compose the list of conjunctions and the values are the components
        by which they must be replaced
    """
    transformations = {}
    for f in list_of_conjunctions:
        c_matrix = args_co_occurence_graph(f, existential_vars)
        components = connected_components(c_matrix)

        if isinstance(f, Quantifier):
            transformations[f] = calc_new_fresh_symbol(f, existential_vars)
            continue

        for c in components:
            form = [f.formulas[i] for i in c]
            if len(form) > 1:
                form = Conjunction(form)
                transformations[form] = \
                    calc_new_fresh_symbol(form, existential_vars)
            else:
                conj = Conjunction(form)
                transformations[form[0]] = \
                    calc_new_fresh_symbol(conj, existential_vars)

    return transformations


def calc_new_fresh_symbol(formula, existential_vars):
    """Given a formula and a list of variables, this function creates a new
    symbol containing only the unbound variables of the formula.

    Parameters
    ----------
    formula : Definition
        Formula to be transformed.

    existential_vars : set
        Set of variables associated with existentials.

    Returns
    -------
    Symbol
        New symbol containing only the unbound variables of the formula.

    """
    cvars = extract_logic_free_variables(formula) - existential_vars

    fresh_symbol = Symbol.fresh()
    new_symbol = fresh_symbol(tuple(cvars))

    return new_symbol


def symbol_connected_components(expression):
    if not isinstance(expression, NaryLogicOperator):
        raise ValueError(
            "Connected components can only be computed "
            "for n-ary logic operators."
        )
    c_matrix = symbol_co_occurence_graph(expression)
    components = connected_components(c_matrix)

    operation = type(expression)
    return [
        operation(tuple(expression.formulas[i] for i in component))
        for component in components
    ]


def symbol_co_occurence_graph(expression):
    """Symbol co-ocurrence graph expressed as
    an adjacency matrix.

    Parameters
    ----------
    expression : NAryLogicExpression
        logic expression for which the adjacency matrix is computed.

    Returns
    -------
    numpy.ndarray
        squared binary array where a component is 1 if there is a
        shared predicate symbol between two subformulas of the
        logic expression.
    """
    c_matrix = np.zeros((len(expression.formulas),) * 2)
    for i, formula in enumerate(expression.formulas):
        atom_symbols = set(a.functor for a in extract_logic_atoms(formula))
        for j, formula_ in enumerate(expression.formulas[i + 1:]):
            atom_symbols_ = set(
                a.functor for a in extract_logic_atoms(formula_)
            )
            if not atom_symbols.isdisjoint(atom_symbols_):
                c_matrix[i, i + 1 + j] = 1
                c_matrix[i + 1 + j, i] = 1
    return c_matrix


def args_co_occurence_graph(expression, variable_to_use=None):
    """Arguments co-ocurrence graph expressed as
    an adjacency matrix.

    Parameters
    ----------
    expression : LogicOperator
        logic expression for which the adjacency matrix is computed.

    Returns
    -------
    numpy.ndarray
        squared binary array where a component is 1 if there is a
        shared argument between two formulas of the
        logic expression.
    """

    if isinstance(expression, Quantifier):
        return np.ones((1,))

    c_matrix = np.zeros((len(expression.formulas),) * 2)
    for i, formula in enumerate(expression.formulas):
        f_args = set(b for a in extract_logic_atoms(formula) for b in a.args)
        if variable_to_use is not None:
            f_args &= variable_to_use
        for j, formula_ in enumerate(expression.formulas[i + 1:]):
            f_args_ = set(
                b for a in extract_logic_atoms(formula_) for b in a.args
            )
            if variable_to_use is not None:
                f_args_ &= variable_to_use
            if not f_args.isdisjoint(f_args_):
                c_matrix[i, i + 1 + j] = 1
                c_matrix[i + 1 + j, i] = 1

    return c_matrix


def connected_components(adjacency_matrix):
    """Connected components of an undirected graph.

    Parameters
    ----------
    adjacency_matrix : numpy.ndarray
        squared array representing the adjacency
        matrix of an undirected graph.

    Returns
    -------
    list of integer sets
        connected components of the graph.
    """
    node_idxs = set(range(adjacency_matrix.shape[0]))
    components = []
    while node_idxs:
        idx = node_idxs.pop()
        component = {idx}
        component_follow = [idx]
        while component_follow:
            idx = component_follow.pop()
            idxs = set(adjacency_matrix[idx].nonzero()[0]) - component
            component |= idxs
            component_follow += idxs
        components.append(component)
        node_idxs -= component
    return components


def disjoint_project(rule_dnf, symbol_table):
    """
    Rule that extends the lifted query processing algorithm to handle
    Block-Independent Disjoint (BID) tables which encode mutual exclusivity
    assumptions on top of the tuple-independent assumption of probabilistic
    tables.

    Modifications to the lifted query processing algorithm that are necessary
    to extend it to BID tables are detailed in section 4.3.1 of [1]_.

    Two variants of the rule exist: one for conjunctive queries, and one for
    disjunctive queries.

    [1] Suciu, Dan, Dan Olteanu, Christopher Ré, and Christoph Koch, eds.
    Probabilistic Databases. Synthesis Lectures on Data Management 16. San
    Rafael, Calif.: Morgan & Claypool Publ, 2011.

    """
    if len(rule_dnf.formulas) == 1:
        conjunctive_query = rule_dnf.formulas[0]
        return disjoint_project_conjunctive_query(
            conjunctive_query, symbol_table
        )
    return disjoint_project_disjunctive_query(rule_dnf, symbol_table)


def disjoint_project_conjunctive_query(conjunctive_query, symbol_table):
    """
    First variant of the disjoint project on a CQ in CNF.

    A disjoint project operator is applied whenever any of the atoms in the CQ
    has only constants in its key positions and has at least one variable in a
    non-key position.

    Note: as variables are removed through shattering, this only applies to
    all probabilistic choice variables.

    """
    free_variables = extract_logic_free_variables(conjunctive_query)
    atoms_with_constants_in_all_key_positions = set(
        atom
        for atom in extract_logic_atoms(conjunctive_query)
        if is_probabilistic_atom_with_constants_in_all_key_positions(
            atom, symbol_table
        )
    )
    if not atoms_with_constants_in_all_key_positions:
        return False, None
    nonkey_variables = set.union(
        *(
            extract_nonkey_variables(atom, symbol_table)
            for atom in atoms_with_constants_in_all_key_positions
        )
    )
    for atom in atoms_with_constants_in_all_key_positions:
        if not isinstance(symbol_table[atom.functor], ProbabilisticChoiceSet):
            raise NeuroLangException(
                "Any atom with constants in all its key positions should be "
                "a probabilistic choice atom"
            )
    conjunctive_query = (
        RemoveExistentialOnVariables(nonkey_variables)
        .walk(conjunctive_query)
    )
    plan = dalvi_suciu_lift(conjunctive_query, symbol_table)
    attributes = tuple(
        str2columnstr_constant(v.name)
        for v in free_variables
    )
    plan = rap.DisjointProjection(plan, attributes)
    return True, plan


def disjoint_project_disjunctive_query(disjunctive_query, symbol_table):
    """
    Second variant of the disjoint project on a UCQ in DNF.

    This rule applies whenever the given query Q can be written as Q = Q1 v Q',
    where the conjunctive query Q1 has an atom where all key attribute are
    constants, and Q' is any other UCQ.

    Then we return P(Q) = P(Q1) + P(Q') - P(Q1 ∧ Q') with subsequent recursive
    calls to the resolution algorithms. Note that a disjoint project should be
    applied during the calculation of P(Q1) and P(Q1 ∧ Q').

    """
    matching_disjuncts = (
        _get_disjuncts_containing_atom_with_all_key_attributes(
            disjunctive_query, symbol_table
        )
    )
    for disjunct in matching_disjuncts:
        has_safe_plan, plan = _apply_disjoint_project_ucq_rule(
            disjunctive_query, disjunct, symbol_table
        )
        if has_safe_plan:
            return has_safe_plan, plan
    return False, None


def _get_disjuncts_containing_atom_with_all_key_attributes(ucq, symbol_table):
    matching_disjuncts = set()
    for disjunct in ucq.formulas:
        if any(
            is_probabilistic_atom_with_constants_in_all_key_positions(
                atom, symbol_table
            )
            for atom in extract_logic_atoms(disjunct)
        ):
            matching_disjuncts.add(disjunct)
    return matching_disjuncts


def _apply_disjoint_project_ucq_rule(
    disjunctive_query: Union,
    disjunct: Conjunction,
    symbol_table: TypedSymbolTableMixin,
):
    free_vars = extract_logic_free_variables(disjunctive_query)
    head = add_existentials_except(disjunct, free_vars)
    head_plan = dalvi_suciu_lift(head, symbol_table)
    if not is_pure_lifted_plan(head_plan):
        return False, None
    tail = set(disjunctive_query.formulas) - {disjunct}
    tail = add_existentials_except(
        Conjunction(tuple(tail)),
        free_vars,
    )
    tail_plan = dalvi_suciu_lift(tail, symbol_table)
    if not is_pure_lifted_plan(tail_plan):
        return False, None
    head_and_tail = add_existentials_except(
        Conjunction(disjunctive_query.formulas), free_vars
    )
    head_and_tail_plan = dalvi_suciu_lift(head_and_tail, symbol_table)
    if not is_pure_lifted_plan(head_and_tail_plan):
        return False, None
    return True, rap.WeightedNaturalJoin(
        (head_plan, tail_plan, head_and_tail_plan),
        (Constant(1), Constant(1), Constant(-1)),
    )


def extract_nonkey_variables(atom, symbol_table):
    """
    Get all variables in the atom that occur on non-key attributes.

    Makes the assumption that the atom is probabilistic.

    As we only support probabilistic choices and not all BID tables, this can
    only be a variable occurring in a probabilistic choice.

    """
    if is_atom_a_probabilistic_choice_relation(atom, symbol_table):
        return {arg for arg in atom.args if isinstance(arg, Symbol)}
    return set()


def has_separator_variables(query, symbol_table):
    """Returns true if `query` has a separator variable and the plan.

    According to Dalvi and Suciu [1]_ if `query` is in DNF,
    a variable z is called a separator variable if Q starts with ∃z,
    that is, Q = ∃z.Q1, for some query expression Q1, and (a) z
    is a root variable (i.e. it appears in every atom),
    (b) for every relation symbol R, there exists an attribute (R, iR)
    such that every atom with symbol R has z in position iR. This is
    equivalent, in datalog syntax, to Q ≡ Q0←∃z.Q1.

    Also, according to Suciu [2]_ the dual is also true,
    if `query` is in CNF i.e. the separation variable z needs to
    be universally quantified, that is Q = ∀x.Q1. But this is not
    implemented.

    [1] Dalvi, N. & Suciu, D. The dichotomy of probabilistic inference
    for unions of conjunctive queries. J. ACM 59, 1–87 (2012).
    [2] Suciu, D. Probabilistic Databases for All. in Proceedings of the
    39th ACM SIGMOD-SIGACT-SIGAI Symposium on Principles of Database Systems
    19–31 (ACM, 2020).

    Parameters
    ----------
    query : LogicExpression
        UCQ to check if it can have a plan based on a separator variable
        strategy.
    symbol_table : dict
        dictionary of symbols and probabilistic/deterministic fact sets.

    Returns
    -------
    boolean, Expression
        Returns true and the plan if the query has separation variables,
        if not, False and None.
    """
    svs, expression = find_separator_variables(query, symbol_table)
    if len(svs) > 0:
        return True, separator_variable_plan(
            expression, svs, symbol_table
        )
    else:
        return False, None


def find_separator_variables(query, symbol_table):
    '''
    According to Dalvi and Suciu [1]_ if `query` is rewritten in prenex
    normal form (PNF) with a DNF matrix, then a variable z is called a
    separator variable if Q starts with ∃z that is, Q = ∃z.Q1, for some
    query expression Q1, and:
      a. z is a root variable (i.e. it appears in every atom); and
      b. for every relation symbol R in Q1, there exists an attribute (R, iR)
      such that every atom with symbol R has z in position iR.

    This algorithm assumes that Q1 can't be splitted into independent
    formulas.

    .. [1] Dalvi, N. & Suciu, D. The dichotomy of probabilistic inference
    for unions of conjunctive queries. J. ACM 59, 1–87 (2012).
    .. [2] Suciu, D. Probabilistic Databases for All. in Proceedings of the
    39th ACM SIGMOD-SIGACT-SIGAI Symposium on Principles of Database Systems
    19–31 (ACM, 2020).
    '''
    exclude_variables = extract_logic_free_variables(query)
    query = unify_existential_variables(query)

    if isinstance(query, NaryLogicOperator):
        formulas = query.formulas
    else:
        formulas = [query]

    candidates = extract_probabilistic_root_variables(formulas, symbol_table)
    all_atoms = extract_logic_atoms(query)

    separator_variables = set()
    for var in candidates:
        atom_positions = {}
        for atom in all_atoms:
            functor = atom.functor
            pos_ = {i for i, v in enumerate(atom.args) if v == var}
            if any(
                pos_.isdisjoint(pos)
                for pos in atom_positions.setdefault(functor, [])
            ):
                break
            atom_positions[functor].append(pos_)
        else:
            separator_variables.add(var)

    query = convert_to_pnf_with_dnf_matrix(query)
    query = PQD.walk(query)
    return separator_variables - exclude_variables, query


def extract_probabilistic_root_variables(formulas, symbol_table):
    candidates = None
    for formula in formulas:
        probabilistic_atoms = OrderedSet(
            atom
            for atom in extract_logic_atoms(formula)
            if not is_atom_a_deterministic_relation(atom, symbol_table)
        )
        if len(probabilistic_atoms) == 0:
            continue
        root_variables = reduce(
            lambda y, x: set(x.args) & y,
            probabilistic_atoms[1:],
            set(probabilistic_atoms[0].args)
        )
        pchoice_atoms = OrderedSet(
            atom
            for atom in probabilistic_atoms
            if is_atom_a_probabilistic_choice_relation(atom, symbol_table)
        )
        # all variables occurring in probabilistic choices cannot occur in key
        # positions, as probabilistic choice relations have no key attribute
        # (making their respective tuples mutually exclusive)
        nonkey_variables = set().union(*(set(a.args) for a in pchoice_atoms))
        # variables occurring in non-key positions cannot be root variables
        # because root variables must occur in every atom in a key position
        root_variables -= nonkey_variables
        if candidates is None:
            candidates = root_variables
        else:
            candidates &= root_variables
    if candidates is None:
        candidates = set()
    return candidates


class IsPureLiftedPlan(PatternWalker):
    @add_match(NonLiftable)
    def non_liftable(self, expression):
        return False

    @add_match(NAryRelationalAlgebraOperation)
    def nary(self, expression):
        return all(
            self.walk(relation)
            for relation in expression.relations
        )

    @add_match(BinaryRelationalAlgebraOperation)
    def binary(self, expression):
        return (
            self.walk(expression.relation_left) &
            self.walk(expression.relation_right)
        )

    @add_match(UnaryRelationalAlgebraOperation)
    def unary(self, expression):
        return self.walk(expression.relation)

    @add_match(...)
    def other(self, expression):
        return True


def is_pure_lifted_plan(query):
    return IsPureLiftedPlan().walk(query)


def separator_variable_plan(expression, separator_variables, symbol_table):
    """Generate RAP plan for the logic query expression assuming it has
    the given separator varialbes.

    Parameters
    ----------
    expression : LogicExpression
        expression to generate the plan for.
    separator_variables : Set[Symbol]
        separator variables for expression.
    symbol_table : mapping of symbol to probabilistic table.
        mapping used to figure out specific strategies according to
        the probabilistic table type.

    Returns
    -------
    RelationalAlgebraOperation
        plan for the logic expression.
    """
    variables_to_project = extract_logic_free_variables(expression)
    if any(
        isinstance(expression, ExistentialPredicate)
        for _, expression in expression_iterator(expression)
    ):
        make_implicit = MakeExistentialsImplicit()
        quantifier_class = ExistentialPredicate
        projection_class = rap.IndependentProjection
    else:
        make_implicit = MakeUniversalsImplicit()
        quantifier_class = UniversalPredicate
        projection_class = rap.IndependentProjectionUniversal

    expression = make_implicit.walk(expression)

    quantified_to_add = (
        extract_logic_free_variables(expression) -
        variables_to_project -
        separator_variables
    )

    for v in quantified_to_add:
        expression = quantifier_class(v, expression)
    return projection_class(
        dalvi_suciu_lift(expression, symbol_table),
        tuple(
            str2columnstr_constant(v.name)
            for v in variables_to_project
        )
    )


def variable_co_occurrence_graph(expression):
    """Free variable co-ocurrence graph expressed as
    an adjacency matrix.

    Parameters
    ----------
    expression : NAryLogicExpression
        logic expression for which the adjacency matrixis computed.

    Returns
    -------
    numpy.ndarray
        squared binary array where a component is 1 if there is a
        shared free variable between two subformulas of the logic expression.
    """
    c_matrix = np.zeros((len(expression.formulas),) * 2)
    for i, formula in enumerate(expression.formulas):
        free_variables = extract_logic_free_variables(formula)
        for j, formula_ in enumerate(expression.formulas[i + 1:]):
            free_variables_ = extract_logic_free_variables(formula_)
            if not free_variables.isdisjoint(free_variables_):
                c_matrix[i, i + 1 + j] = 1
                c_matrix[i + 1 + j, i] = 1
    return c_matrix


def components_plan(
    components, operation, symbol_table,
    negative_operation=None,
    unary_negative_operation=None,
):
    positive_formulas = []
    negative_formulas = []
    for component in components:
        component = RTO.walk(component)
        if isinstance(component, Negation):
            formula = dalvi_suciu_lift(component.formula, symbol_table)
            negative_formulas.append(formula)
        else:

            formula = dalvi_suciu_lift(component, symbol_table)
            positive_formulas.append(formula)

    if len(positive_formulas) > 0:
        output = reduce(operation, positive_formulas[1:], positive_formulas[0])

    if (
        len(positive_formulas) > 0 and
        len(negative_formulas) > 0
    ):
        if negative_operation is None:
            raise ValueError(
                "If negative components are included,"
                " a negative operation should be provided"
            )
        output = reduce(negative_operation, negative_formulas, output)
    elif len(negative_formulas) > 0:
        if unary_negative_operation is None:
            raise ValueError(
                "If negative components are included,"
                " and no positive formulas exist,"
                " a unary negative operation should be provided"
            )
        complemented_negative_formulas = [
            unary_negative_operation(f) for f in negative_formulas
        ]
        output = reduce(
            operation,
            complemented_negative_formulas[1:],
            complemented_negative_formulas[0]
        )
    return output


def inclusion_exclusion_conjunction(expression, symbol_table):
    """Produce a RAP query plan for the conjunction logic query
    based on the inclusion exclusion formula.

    Parameters
    ----------
    expression : Conjunction
        logic expression to produce a plan for.
    symbol_table : mapping of symbol to probabilistic table.
        mapping used to figure out specific strategies according to
        the probabilistic table type.

    Returns
    -------
    RelationalAlgebraOperation
        plan for the logic expression.
    """
    formula_powerset = []
    for formula in powerset(expression.formulas):
        if len(formula) == 0:
            continue

        formula = RTO.walk(convert_ucq_to_ccq(
            Disjunction(tuple(formula)), transformation="DNF")
        )
        formula_powerset.append(formula)
    formulas_weights = _formulas_weights(formula_powerset)
    new_formulas, weights = zip(*(
        (dalvi_suciu_lift(formula, symbol_table), weight)
        for formula, weight in formulas_weights.items()
        if weight != 0
    ))

    return rap.WeightedNaturalJoin(
        tuple(new_formulas),
        tuple(Constant[int](w) for w in weights)
    )


def _formulas_weights(formula_powerset):
    # Using list set instead of a dictionary
    # due to a strange bug in dictionary lookups with Expressions
    # as keys
    formula_containments = [
        set()
        for formula in formula_powerset
    ]

    for i, f0 in enumerate(formula_powerset):
        for j, f1 in enumerate(formula_powerset[i + 1:], start=i + 1):
            for i0, c0, i1, c1 in ((i, f0, j, f1), (j, f1, i, f0)):
                if (
                    c0 not in formula_containments[i1] and
                    is_contained(c0, c1)
                ):
                    formula_containments[i0] |= (
                        {c1} | formula_containments[i1] - {c0}
                    )

    fcs = {
        formula: containment
        for formula, containment in
        zip(formula_powerset, formula_containments)
    }
    formulas_weights = mobius_weights(fcs)
    return formulas_weights


def mobius_weights(formula_containments):
    _mobius_weights = {}
    for formula in formula_containments:
        _mobius_weights[formula] = mobius_function(
            formula, formula_containments, _mobius_weights
        )
    return _mobius_weights


def mobius_function(formula, formula_containments, known_weights=None):
    if known_weights is None:
        known_weights = dict()
    if formula in known_weights:
        return known_weights[formula]
    res = -1
    for f in formula_containments[formula]:
        weight = known_weights.setdefault(
            f,
            mobius_function(
                f,
                formula_containments, known_weights=known_weights
            )
        )
        res += weight
    return -res


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))


def is_probabilistic_atom_with_constants_in_all_key_positions(
    atom, symbol_table
):
    """
    As we only handle probabilistic choice relations (and not more general BID
    tables), which are relations with no key attribute, there are only two
    cases:
    (1) if the atom is a probabilistic choice, then it validates the
    requirement of having constants in all key positions (it has no key
    attribute), or
    (2) if the atom is a tuple-independent relation, all its attributes are key
    attributes, and so all its terms must be constants for the requirement to
    be validated.

    """
    return is_atom_a_probabilistic_choice_relation(atom, symbol_table) or (
        not is_atom_a_deterministic_relation(atom, symbol_table)
        and (
            isinstance(atom.functor, Symbol)
            and atom.functor in symbol_table
            and isinstance(symbol_table[atom.functor], ProbabilisticFactSet)
            and all(isinstance(arg, Constant) for arg in atom.args)
        )
    )



import json
from concurrent.futures import Future
from typing import AbstractSet, Tuple
from unittest.mock import create_autospec
from uuid import uuid4

import pytest
from neurolang.exceptions import NeuroLangException
from neurolang.type_system import Unknown
from neurolang.utils.relational_algebra_set import (
    NamedRelationalAlgebraFrozenSet,
)
from ..responses import (
    CustomQueryResultsEncoder,
    QueryResults,
)


@pytest.fixture
def future():
    future = create_autospec(Future)
    future.cancelled.return_value = False
    future.done.return_value = False
    future.running.return_value = True
    return future


@pytest.fixture
def error():
    return NeuroLangException("Something went wrong")


@pytest.fixture
def data():
    data = [
        (5, "dog", frozenset({(1, 2), (5, 6)})),
        (10, "cat", frozenset({(5, 6), (8, 9)})),
        (-5.25, "mouse", frozenset({(8, 9), (12, 13)})),
    ]
    return data


@pytest.fixture
def result(data):
    ans = NamedRelationalAlgebraFrozenSet(("a", "b", "c"), data)
    ans.row_type = Tuple[float, str, AbstractSet[Unknown]]
    results = {"ans": ans}
    return results


@pytest.fixture
def results(data):
    ans = NamedRelationalAlgebraFrozenSet(("a", "b", "c"), data)
    ans.row_type = Tuple[float, str, AbstractSet[Unknown]]
    voxels = NamedRelationalAlgebraFrozenSet(
        ("i", "j", "k"), [[12, 15, 98], [107, 2, 33], [89, 8, 34]]
    )
    voxels.row_type = Tuple[int, int, int]
    results = {"ans": ans, "Voxel": voxels}
    return results


def test_query_results_has_metadata(future):
    uuid = str(uuid4())
    page = 2
    limit = 100
    qr = QueryResults(uuid, future, page, limit)
    assert qr.uuid == uuid
    assert qr.cancelled == False
    assert qr.done == False
    assert qr.running == True


def test_query_results_has_exceptions(future, error):
    future.done.return_value = True
    future.exception.return_value = error

    qr = QueryResults(str(uuid4()), future)
    assert qr.errorName == str(type(error))
    assert qr.errorDoc == error.__doc__
    assert qr.message == str(error)


def test_query_results_has_results(future, results, data):
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = results

    qr = QueryResults(str(uuid4()), future, get_values=True)
    assert qr.results is not None
    assert qr.results["ans"]["row_type"] == [
        str(t) for t in (float, str, AbstractSet[Unknown])
    ]
    assert qr.results["ans"]["columns"] == ["a", "b", "c"]
    assert qr.results["ans"]["size"] == 3
    assert qr.results["ans"]["values"] == [[a, b, str(c)] for a, b, c in data]
    assert qr.results["Voxel"]["row_type"] == [str(t) for t in (int, int, int)]
    assert qr.results["Voxel"]["columns"] == ["i", "j", "k"]
    assert qr.results["Voxel"]["size"] == 3


def test_query_results_does_not_return_values_by_default(future, result):
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = result

    qr = QueryResults(str(uuid4()), future)
    assert qr.results is not None
    assert qr.results["ans"]["row_type"] == [
        str(t) for t in (float, str, AbstractSet[Unknown])
    ]
    assert qr.results["ans"]["columns"] == ["a", "b", "c"]
    assert qr.results["ans"]["size"] == 3
    assert "values" not in qr.results["ans"]


def test_query_results_only_returns_requested_symbol(future, results, data):
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = results

    qr = QueryResults(str(uuid4()), future, "Voxel")
    assert qr.results is not None
    assert len(qr.results) == 1
    assert qr.results["Voxel"]["row_type"] == [str(t) for t in (int, int, int)]
    assert qr.results["Voxel"]["columns"] == ["i", "j", "k"]
    assert qr.results["Voxel"]["size"] == 3
    assert len(qr.results["Voxel"]["values"]) == 3

    qr = QueryResults(str(uuid4()), future, "ans")
    assert qr.results is not None
    assert len(qr.results) == 1
    assert qr.results["ans"]["row_type"] == [
        str(t) for t in (float, str, AbstractSet[Unknown])
    ]
    assert qr.results["ans"]["columns"] == ["a", "b", "c"]
    assert qr.results["ans"]["size"] == 3
    assert qr.results["ans"]["values"] == [[a, b, str(c)] for a, b, c in data]


def test_query_results_can_paginate(future, result, data):
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = result

    qr = QueryResults(str(uuid4()), future, start=0, length=1, get_values=True)
    assert qr.results is not None
    assert qr.results["ans"]["values"] == [
        [a, b, str(c)] for a, b, c in data[0:1]
    ]

    qr = QueryResults(str(uuid4()), future, start=1, length=2, get_values=True)
    assert qr.results is not None
    assert qr.results["ans"]["values"] == [
        [a, b, str(c)] for a, b, c in data[1:3]
    ]

    qr = QueryResults(str(uuid4()), future, start=2, length=2, get_values=True)
    assert qr.results is not None
    assert qr.results["ans"]["values"] == [
        [a, b, str(c)] for a, b, c in data[2:]
    ]


def test_query_results_can_sort(future, result, data):
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = result

    qr = QueryResults(
        str(uuid4()), future, start=0, length=2, sort=1, get_values=True
    )
    assert qr.results is not None
    assert qr.results["ans"]["values"] == [
        [a, b, str(c)] for a, b, c in data[1::-1]
    ]


def test_query_results_can_serialize_to_json(future, result, data):
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = result

    uuid = str(uuid4())
    qr = QueryResults(uuid, future, get_values=True)
    json_string = json.dumps(qr, cls=CustomQueryResultsEncoder)
    expected = {
        "uuid": uuid,
        "cancelled": False,
        "running": True,
        "done": True,
        "start": 0,
        "length": 50,
        "asc": True,
        "sort": -1,
        "results": {
            "ans": {
                "columns": ["a", "b", "c"],
                'last_parsed_symbol': False,
                'probabilistic': False,
                "row_type": [
                    str(t) for t in (float, str, AbstractSet[Unknown])
                ],
                "size": 3,
                "values": [[a, b, str(c)] for a, b, c in data],
            }
        },
    }
    assert json.loads(json_string) == expected



from ..exceptions import NotInFONegE
from ..expression_walker import (
    ChainedWalker,
    ExpressionWalker,
    IdentityWalker,
    PatternWalker,
    ReplaceExpressionWalker,
    ReplaceSymbolWalker,
    add_match
)
from ..logic.expression_processing import (
    ExtractFreeVariablesWalker,
    WalkLogicProgramAggregatingSets,
    extract_logic_free_variables
)
from ..utils.orderedset import OrderedSet
from . import (
    Conjunction,
    Constant,
    Disjunction,
    ExistentialPredicate,
    FunctionApplication,
    Implication,
    NaryLogicOperator,
    Negation,
    Quantifier,
    Symbol,
    UnaryLogicOperator,
    UniversalPredicate
)


class LogicExpressionWalker(PatternWalker):
    @add_match(Quantifier)
    def walk_quantifier(self, expression):
        new_head = self.walk(expression.head)
        new_body = self.walk(expression.body)
        if new_head != expression.head or new_body != expression.body:
            return self.walk(expression.apply(new_head, new_body))
        else:
            return expression

    @add_match(Constant)
    def walk_constant(self, expression):
        return expression

    @add_match(Symbol)
    def walk_symbol(self, expression):
        return expression

    @add_match(UnaryLogicOperator)
    def walk_negation(self, expression):
        formula = self.walk(expression.formula)
        if formula != expression.formula:
            return self.walk(expression.apply(self.walk(expression.formula)))
        else:
            return expression

    @add_match(FunctionApplication)
    def walk_function(self, expression):
        new_args = tuple(map(self.walk, expression.args))
        if new_args != expression.args:
            return self.walk(expression.apply(
                    expression.functor, new_args
                ))
        else:
            return expression

    @add_match(NaryLogicOperator)
    def walk_nary(self, expression):
        new_formulas = tuple(self.walk(expression.formulas))
        if any(nf != f for nf, f in zip(new_formulas, expression.formulas)):
            return self.walk(expression.apply(new_formulas))
        else:
            return expression

    @add_match(Implication)
    def walk_implication(self, expression):
        antecedent = self.walk(expression.antecedent)
        consequent = self.walk(expression.consequent)

        if (
            antecedent != expression.antecedent or
            consequent != expression.consequent
        ):
            return self.walk(Implication(consequent, antecedent))
        else:
            return expression


class EliminateImplications(LogicExpressionWalker):
    """
    Removes the implication ocurrences of an expression.
    """

    @add_match(Implication(..., ...))
    def remove_implication(self, implication):
        c = self.walk(implication.consequent)
        a = self.walk(implication.antecedent)
        return self.walk(Disjunction((c, Negation(a))))


class RemoveTrivialOperationsMixin(PatternWalker):
    @add_match(NaryLogicOperator, lambda e: len(e.formulas) == 1)
    def remove_single(self, expression):
        return self.walk(expression.formulas[0])

    @add_match(Negation(Negation(...)))
    def remove_double_negation(self, expression):
        return self.walk(expression.formula.formula)


class MoveNegationsToAtomsSimpleOperationsMixin(PatternWalker):
    """
    Moves the negations the furthest possible to the atoms. On
    conjunction, disjunction, and negation operations.
    """
    @add_match(Negation(Conjunction(...)))
    def negated_conjunction(self, negation):
        conj = negation.formula
        formulas = map(lambda e: self.walk(Negation(e)), conj.formulas)
        return self.walk(Disjunction(tuple(formulas)))

    @add_match(Negation(Disjunction(...)))
    def negated_disjunction(self, negation):
        disj = negation.formula
        formulas = map(lambda e: self.walk(Negation(e)), disj.formulas)
        return self.walk(Conjunction(tuple(formulas)))


class PullUniversalUpFromNegation(PatternWalker):
    @add_match(Negation(UniversalPredicate(..., ...)))
    def negated_universal(self, negation):
        quantifier = negation.formula
        x = quantifier.head
        return self.walk(ExistentialPredicate(x, Negation(quantifier.body)))


class PullExistentialUpFromNegation(PatternWalker):
    @add_match(Negation(ExistentialPredicate(..., ...)))
    def negated_existential(self, negation):
        quantifier = negation.formula
        x = quantifier.head
        return self.walk(UniversalPredicate(x, Negation(quantifier.body)))


class MoveNegationsToAtomsMixin(
    RemoveTrivialOperationsMixin,
    PullExistentialUpFromNegation,
    PullUniversalUpFromNegation,
    MoveNegationsToAtomsSimpleOperationsMixin,
    PatternWalker
):
    """
    Moves the negations the furthest possible to the atoms.
    Assumes that there are no implications in the expression.
    """
    pass


class MoveNegationsToAtoms(
    MoveNegationsToAtomsMixin,
    LogicExpressionWalker
):
    pass


class FONegELogicExpression(LogicExpressionWalker):
    """
    Walk a logic expression verifying that it is only in
    FO Neg Existential.
    """
    @add_match(UniversalPredicate)
    def abort_universal_predicate(self, expression):
        raise NotInFONegE(
            f"Forumla {expression} not in FO Existental Negation"
        )

    @add_match(Implication)
    def abort_implication(self, expression):
        raise NotInFONegE(
            f"Forumla {expression} not in FO Existental Negation"
        )


class MoveNegationsToAtomsInFONegE(
    MoveNegationsToAtomsSimpleOperationsMixin,
    PullUniversalUpFromNegation,
    FONegELogicExpression
):
    """
    Moves the negations the furthest possible to the atoms.
    Assumes that there are no implications in the expression.
    Expressions assumed to be in FO with existential and negation
    only.
    """
    pass


class FactorQuantifiersMixin(PatternWalker):
    """
    Factor quantifiers up on conjunctions and disjunctions.
    """
    @add_match(
        Disjunction,
        lambda exp: any(isinstance(f, Quantifier) for f in exp.formulas),
    )
    def disjunction_with_quantifiers(self, expression):
        expression = Disjunction(tuple(
            self.walk(formula)
            for formula in expression.formulas
        ))
        formulas, quantifiers = \
            self._extract_quantifiers_freshen_variables(expression)
        exp = Disjunction(tuple(formulas))
        for q in reversed(quantifiers):
            exp = q.apply(q.head, exp)
        return self.walk(exp)

    @add_match(
        Conjunction,
        lambda exp: any(isinstance(f, Quantifier) for f in exp.formulas),
    )
    def conjunction_with_quantifiers(self, expression):
        expression = Conjunction(tuple(
            self.walk(formula)
            for formula in expression.formulas
        ))
        formulas, quantifiers = \
            self._extract_quantifiers_freshen_variables(expression)
        exp = Conjunction(tuple(formulas))
        for q in reversed(quantifiers):
            exp = q.apply(q.head, exp)
        return self.walk(exp)

    @staticmethod
    def _extract_quantifiers_freshen_variables(expression):
        formulas = []
        quantifiers = []
        variables = extract_logic_free_variables(expression)
        for f in expression.formulas:
            while isinstance(f, Quantifier):
                if f.head in variables:
                    f = ReplaceSymbolWalker(
                        {f.head: Symbol[f.type].fresh()}
                    ).walk(f)
                variables.add(f.head)
                quantifiers.append(f)
                f = f.body
            formulas.append(f)
        return formulas, quantifiers


class MoveQuantifiersUp(
    FactorQuantifiersMixin,
    PullUniversalUpFromNegation,
    PullExistentialUpFromNegation,
    LogicExpressionWalker
):
    """
    Moves the quantifiers up in order to format the expression
    in prenex normal form. Assumes the expression contains no implications
    and the variables of the quantifiers are not repeated.
    """
    pass


class MoveQuantifiersUpFONegE(FactorQuantifiersMixin, FONegELogicExpression):
    """
    Moves the quantifiers up in order to format the expression
    in prenex normal form. Assumes the expression contains no implications,
    that the variables of the quantifiers are not repeated, and that
    the query is in FO with Negation Existential as only quantifier.
    """
    pass


class DesambiguateQuantifiedVariables(LogicExpressionWalker):
    """
    Replaces each quantified variale to a fresh one.
    """
    @add_match(
        NaryLogicOperator,
        lambda expression: len(expression.formulas) > 1
    )
    def nary_logic_operator(self, expression):
        free_variables = extract_logic_free_variables(expression)
        expression_first = self.walk(expression.formulas[0])
        bound_variables_first = ExtractBoundVariables().walk(expression_first)

        new_expression_tail = expression.apply(tuple(
            FreshenVariablesWhenQuantified(
                free_variables | bound_variables_first
            ).walk(formula)
            for formula in expression.formulas[1:]
        ))
        new_expression_tail = self.walk(new_expression_tail)

        new_expression = expression.apply(
            (expression_first,) + new_expression_tail.formulas
        )

        return new_expression

    @add_match(Quantifier)
    def quantifier(self, expression):
        return expression.apply(
            expression.head,
            self.walk(expression.body)
        )

    @add_match(Implication)
    def implication(self, expression):
        convert_to_disjunction = Disjunction(
            (Negation(expression.antecedent), expression.consequent)
        )
        walked_disjunction = self.walk(convert_to_disjunction)

        if (
            isinstance(walked_disjunction, Disjunction) and
            len(walked_disjunction.formulas) == 2 and
            isinstance(walked_disjunction.formulas[0], Negation)
        ):
            return Implication(
                walked_disjunction.formulas[1],
                walked_disjunction.formulas[0].formula
            )
        else:
            return walked_disjunction


class FreshenVariablesWhenQuantified(LogicExpressionWalker):
    def __init__(self, free_variables):
        self.free_variables = free_variables

    @add_match(Quantifier)
    def replace_variable_in_quantifier(self, expression):
        head, body = expression.unapply()
        if head in self.free_variables:
            new_head = head.fresh()
            new_body = (
                ReplaceExpressionWalker({expression.head: new_head})
                .walk(expression.body)
            )
            head = new_head
            body = new_body

        new_expression = expression.apply(
            head,
            FreshenVariablesWhenQuantified({head}).walk(body)
        )
        return new_expression

    @add_match(Implication)
    def walk_through_implication(self, expression):
        return Implication(
            self.walk(expression.consequent),
            self.walk(expression.antecedent)
        )


class ExtractBoundVariables(WalkLogicProgramAggregatingSets):
    @add_match(Quantifier)
    def quantifier(self, expression):
        return OrderedSet([expression.head]) | self.walk(expression.body)

    @add_match(FunctionApplication)
    def process_function_application(self, expression):
        return OrderedSet([])

    @add_match(Symbol)
    def process_symbol(self, expression):
        return OrderedSet([])


class ReplaceFreeSymbolWalker(ReplaceSymbolWalker):
    @add_match(Quantifier)
    def stop_if_bound(self, expression):
        s = expression.head
        r = self.symbol_replacements.pop(s, None)
        expression.body = self.walk(expression.body)
        if r:
            self.symbol_replacements[s] = r
        return expression


class UsedQuantifiers(PatternWalker):
    @add_match(FunctionApplication)
    def function(self, exp):
        return set()

    @add_match(Symbol)
    def symbol(self, exp):
        return set()

    @add_match(Negation)
    def negation(self, exp):
        return self.walk(exp.formula)

    @add_match(Disjunction)
    def disjunction(self, exp):
        return set.union(*self.walk(exp.formulas))

    @add_match(Implication)
    def implication(self, exp):
        return set.union(self.walk(exp.antecedent), self.walk(exp.consequent))

    @add_match(Conjunction)
    def conjunction(self, exp):
        return set.union(*self.walk(exp.formulas))

    @add_match(Quantifier)
    def quantifier(self, exp):
        return {exp} | self.walk(exp.body)


class DistributeDisjunctions(LogicExpressionWalker):
    @add_match(Disjunction, lambda e: len(e.formulas) > 2)
    def split(self, expression):
        head, *rest = expression.formulas
        rest = self.walk(Disjunction(tuple(rest)))
        new_exp = Disjunction((head, rest))
        return self.walk(new_exp)

    @add_match(Disjunction((..., Conjunction)))
    def rotate(self, expression):
        q, c = expression.formulas
        return self.walk(
            Conjunction(tuple(map(lambda p: Disjunction((q, p)), c.formulas)))
        )

    @add_match(Disjunction((Conjunction, ...)))
    def distribute(self, expression):
        c, q = expression.formulas
        return self.walk(
            Conjunction(tuple(map(lambda p: Disjunction((p, q)), c.formulas)))
        )


class DistributeConjunctions(LogicExpressionWalker):
    @add_match(Conjunction, lambda e: len(e.formulas) > 2)
    def split(self, expression):
        head, *rest = expression.formulas
        rest = self.walk(Conjunction(tuple(rest)))
        new_exp = Conjunction((head, rest))
        return self.walk(new_exp)

    @add_match(Conjunction((..., Disjunction)))
    def rotate(self, expression):
        q, c = expression.formulas
        return self.walk(
            Disjunction(tuple(map(lambda p: Conjunction((q, p)), c.formulas)))
        )

    @add_match(Conjunction((Disjunction, ...)))
    def distribute(self, expression):
        c, q = expression.formulas
        return self.walk(
            Disjunction(tuple(map(lambda p: Conjunction((p, q)), c.formulas)))
        )


class CollapseDisjunctionsMixin(PatternWalker):
    @add_match(
        Disjunction,
        lambda e: any(isinstance(f, Disjunction) for f in e.formulas),
    )
    def disjunction(self, e):
        new_arg = []
        for f in map(self.walk, e.formulas):
            if isinstance(f, Disjunction):
                new_arg.extend(f.formulas)
            else:
                new_arg.append(f)
        return self.walk(Disjunction(tuple(new_arg)))


class CollapseDisjunctions(CollapseDisjunctionsMixin, ExpressionWalker):
    pass


class CollapseConjunctionsMixin(PatternWalker):
    @add_match(
        Conjunction,
        lambda e: any(isinstance(f, Conjunction) for f in e.formulas),
    )
    def conjunction(self, e):
        new_arg = []
        for f in map(self.walk, e.formulas):
            if isinstance(f, Conjunction):
                new_arg.extend(f.formulas)
            else:
                new_arg.append(f)
        return self.walk(Conjunction(tuple(new_arg)))


class CollapseConjunctions(CollapseConjunctionsMixin, ExpressionWalker):
    pass


class RemoveUniversalPredicates(LogicExpressionWalker):
    """
    Changes the universal predicates by equivalent expressions
    using existential quantifiers.
    """

    @add_match(UniversalPredicate)
    def universal(self, expression):
        return self.walk(
            Negation(
                ExistentialPredicate(
                    expression.head, Negation(self.walk(expression.body))
                )
            )
        )


class MakeExistentialsImplicit(LogicExpressionWalker):
    @add_match(ExistentialPredicate)
    def existential(self, expression):
        return self.walk(expression.body)


class MakeUniversalsImplicit(LogicExpressionWalker):
    @add_match(UniversalPredicate)
    def universal(self, expression):
        return self.walk(expression.body)


class RemoveExistentialOnVariables(LogicExpressionWalker):
    def __init__(self, variables_to_eliminate):
        self._variables_to_eliminate = variables_to_eliminate

    @add_match(ExistentialPredicate)
    def existential(self, expression):
        body = self.walk(expression.body)
        if expression.head in self._variables_to_eliminate:
            return body
        else:
            return ExistentialPredicate(expression.head, body)


def convert_to_pnf_with_cnf_matrix(expression):
    walker = ChainedWalker(
        EliminateImplications,
        MoveNegationsToAtoms,
        DesambiguateQuantifiedVariables,
        MoveQuantifiersUp,
        RemoveTrivialOperations,
        DistributeDisjunctions,
        CollapseDisjunctions,
        CollapseConjunctions,
    )
    return walker.walk(expression)


def convert_to_pnf_with_dnf_matrix(expression):
    walker = ChainedWalker(
        EliminateImplications,
        MoveNegationsToAtoms,
        DesambiguateQuantifiedVariables,
        MoveQuantifiersUp,
        RemoveTrivialOperations,
        DistributeConjunctions,
        CollapseDisjunctions,
        CollapseConjunctions,
    )
    return walker.walk(expression)


class ExtractFOLFreeVariables(ExtractFreeVariablesWalker):
    @add_match(Implication)
    def extract_variables_s(self, exp):
        return self.walk(exp.consequent) | self.walk(exp.antecedent)


class DistributeUniversalQuantifiers(PatternWalker):
    @add_match(UniversalPredicate(..., Conjunction))
    def distribute_universal_quantifier(self, uq):
        return self.walk(
            Conjunction(
                tuple(map(self._apply_quantifier(uq.head), uq.body.formulas))
            )
        )

    def _apply_quantifier(self, var):
        def foo(exp):
            fv = ExtractFOLFreeVariables().walk(exp)
            if var in fv:
                exp = UniversalPredicate(var, exp)
            return exp

        return foo


class DistributeImplicationsWithConjunctiveHeads(PatternWalker):
    @add_match(Implication(Conjunction, ...))
    def distribute_implication_with_conjunctive_head(self, impl):
        return self.walk(
            Conjunction(
                tuple(
                    Implication(h, impl.antecedent)
                    for h in impl.consequent.formulas
                )
            )
        )


class RemoveTrivialOperations(
    RemoveTrivialOperationsMixin, LogicExpressionWalker
):
    pass


class PushExistentialsDownMixin(
    CollapseConjunctionsMixin, CollapseDisjunctionsMixin,
    PatternWalker
):
    @add_match(
        ExistentialPredicate,
        lambda expression: (
            expression.head not in
            extract_logic_free_variables(expression.body)
        )
    )
    def remove_trivial_existential(self, expression):
        return self.walk(expression.body)

    @add_match(
        ExistentialPredicate(..., NaryLogicOperator),
        lambda e: len(e.body.formulas) == 1
    )
    def push_eliminate_trivial_operation(self, expression):
        return self.walk(
            ExistentialPredicate(expression.head, expression.body.formulas[0])
        )

    @add_match(ExistentialPredicate(..., Disjunction))
    def push_existential_down_disjunction(self, expression):
        variable = expression.head
        changed = False
        new_formulas = tuple()
        for formula in expression.body.formulas:
            if variable in extract_logic_free_variables(formula):
                changed = True
                formula = self.walk(ExistentialPredicate(variable, formula))
            new_formulas += (formula,)
        if changed:
            res = self.walk(Disjunction(new_formulas))
        else:
            res = expression
        return res

    @add_match(
        ExistentialPredicate(..., Conjunction),
        lambda expression: any(
            isinstance(formula, Negation) and
            expression.head in extract_logic_free_variables(formula)
            for formula in expression.body.formulas
        )
    )
    def dont_push_when_it_can_be_unsafe(self, expression):
        variable = expression.head
        body_formulas = tuple(
            self.walk(formula)
            for formula in expression.body.formulas
        )
        in_ = tuple()
        out_ = tuple()
        negative_logic_free_variables = set()
        for formula in body_formulas:
            if isinstance(formula, Negation):
                negative_logic_free_variables |= extract_logic_free_variables(
                    formula
                )

        for formula in body_formulas:
            if (
                negative_logic_free_variables &
                extract_logic_free_variables(formula)
            ):
                in_ += (formula,)
            else:
                out_ += (formula,)

        if len(out_) == 0:
            res = ExistentialPredicate(variable, self.walk(Conjunction(in_)))
        else:
            res = self.walk(
                Conjunction((
                    ExistentialPredicate(variable, Conjunction(in_)),
                    Conjunction(out_)
                ))
            )
        return res

    @add_match(ExistentialPredicate(..., Conjunction))
    def push_existential_down(self, expression):
        variable = expression.head
        in_ = tuple()
        out_ = tuple()
        for formula in expression.body.formulas:
            if variable in extract_logic_free_variables(formula):
                in_ += (formula,)
            else:
                out_ += (formula,)

        if len(out_) == 0:
            res = ExistentialPredicate(variable, self.walk(Conjunction(in_)))
        elif len(in_) == 0:
            res = self.walk(Conjunction(out_))
        if len(in_) > 0 and len(out_) > 0:
            res = self.walk(
                Conjunction(
                    (
                        ExistentialPredicate(variable, Conjunction(in_)),
                    ) + out_
                )
            )
        return res

    @add_match(ExistentialPredicate(..., ExistentialPredicate))
    def nested_existential(self, expression):
        outer_var = expression.head
        new_body = self.walk(expression.body)
        if new_body != expression.body:
            expression = self.walk(
                ExistentialPredicate(outer_var, new_body)
            )
        else:
            inner_var = expression.body.head
            inner_body = expression.body.body
            swapped_body = ExistentialPredicate(
                outer_var, inner_body
            )
            new_body = self.walk(swapped_body)
            if new_body != swapped_body:
                expression = self.walk(
                    ExistentialPredicate(inner_var, new_body)
                )
        return expression


class PushUniversalsDownMixin(
    CollapseConjunctionsMixin, CollapseDisjunctionsMixin,
    PatternWalker
):
    @add_match(
        UniversalPredicate,
        lambda expression: (
            expression.head not in
            extract_logic_free_variables(expression.body)
        )
    )
    def remove_trivial_universal(self, expression):
        return self.walk(expression.body)

    @add_match(
        UniversalPredicate(..., NaryLogicOperator),
        lambda e: len(e.body.formulas) == 1
    )
    def push_eliminate_trivial_operation(self, expression):
        return self.walk(
            expression.apply(expression.head, expression.body.formulas[0])
        )

    @add_match(UniversalPredicate(..., Conjunction))
    def push_universal_down_conjunction(self, expression):
        variable = expression.head
        changed = False
        new_formulas = tuple()
        for formula in expression.body.formulas:
            if variable in extract_logic_free_variables(formula):
                changed = True
                formula = self.walk(UniversalPredicate(variable, formula))
            new_formulas += (formula,)
        if changed:
            res = self.walk(Conjunction(new_formulas))
        else:
            res = expression
        return res

    @add_match(
        UniversalPredicate(..., Disjunction),
        lambda expression: any(
            expression.head not in extract_logic_free_variables(formula)
            for formula in expression.body.formulas
        )
    )
    def push_universal_down_disjunction(self, expression):
        variable = expression.head
        in_ = tuple()
        out_ = tuple()
        for formula in expression.body.formulas:
            if variable in extract_logic_free_variables(formula):
                in_ += (formula,)
            else:
                out_ += (formula,)

        if len(out_) == 0:
            res = UniversalPredicate(variable, self.walk(Conjunction(in_)))
        elif len(in_) == 0:
            res = self.walk(Disjunction(out_))
        if len(in_) > 0 and len(out_) > 0:
            res = self.walk(
                Disjunction(
                    (
                        UniversalPredicate(variable, Disjunction(in_)),
                    ) + out_
                )
            )
        return res


class PushExistentialsDown(PushExistentialsDownMixin, LogicExpressionWalker):
    pass


class PushUniversalsDown(PushUniversalsDownMixin, LogicExpressionWalker):
    pass


class PushQuantifiersDown(
    RemoveTrivialOperationsMixin,
    PushExistentialsDownMixin, PushUniversalsDownMixin,
    LogicExpressionWalker
):
    pass


class GuaranteeConjunction(IdentityWalker):
    @add_match(..., lambda e: not isinstance(e, Conjunction))
    def guarantee_conjunction(self, expression):
        return Conjunction((expression,))


class GuaranteeDisjunction(IdentityWalker):
    @add_match(..., lambda e: not isinstance(e, Disjunction))
    def guarantee_conjunction(self, expression):
        return Disjunction((expression,))


def nary_op_has_duplicated_formulas(nary_op: NaryLogicOperator) -> bool:
    seen = set()
    for formula in nary_op.formulas:
        if formula in seen:
            return True
        seen.add(formula)
    return False


class RemoveDuplicatedConjunctsDisjuncts(LogicExpressionWalker):
    @add_match(Disjunction, nary_op_has_duplicated_formulas)
    def disjunction(self, disjunction: Disjunction) -> Disjunction:
        return self.walk(self._nary_op_without_duplicates(disjunction))

    @add_match(Conjunction, nary_op_has_duplicated_formulas)
    def conjunction(self, conjunction: Conjunction) -> Conjunction:
        return self.walk(self._nary_op_without_duplicates(conjunction))

    @staticmethod
    def _nary_op_without_duplicates(
        nary_op: NaryLogicOperator,
    ) -> NaryLogicOperator:
        return nary_op.apply(tuple(set(nary_op.formulas)))


class CheckConjunctiveQueryWithNegation(LogicExpressionWalker):
    @add_match(Conjunction)
    def conjunction(self, expression):
        for f in expression.formulas:
            if isinstance(f, Conjunction) or not self.walk(f):
                return False

        return True

    @add_match(FunctionApplication)
    def f_app(self, expression):
        return True

    @add_match(Quantifier)
    def existential_predicate(self, expression):
        if isinstance(expression.body, Conjunction):
            return False

        return self.walk(expression.body)

    @add_match(Negation(FunctionApplication))
    def negation(self, expression):
        return True

    @add_match(...)
    def default(self, expression):
        return False


class ExtractConjunctiveQueryWithNegation(WalkLogicProgramAggregatingSets):
    @add_match(Conjunction, CheckConjunctiveQueryWithNegation().walk)
    def pure_conjunction(self, expression):
        return OrderedSet([expression])

    @add_match(NaryLogicOperator)
    def conjunction(self, expression):
        conjunctions = OrderedSet([])
        for f in expression.formulas:
            conjunctions |= self.walk(f)

        return conjunctions

    @add_match(Quantifier)
    def existential_predicate(self, expression):
        body = self.walk(expression.body)
        if body:
            return OrderedSet([expression])

        return OrderedSet([])

    @add_match(FunctionApplication)
    def f_app(self, expression):
        return OrderedSet([])

    @add_match(Negation)
    def neg(self, expression):
        return self.walk(expression.formula)


class RemoveExistentialPredicates(LogicExpressionWalker):
    @add_match(ExistentialPredicate)
    def existential_predicate(self, existential_predicate):
        return self.walk(existential_predicate.body)



from ..exceptions import ForbiddenExpressionError


class AggregatedVariableReplacedByConstantError(ForbiddenExpressionError):
    pass


class InvalidMagicSetError(ForbiddenExpressionError):
    """
    Generic class for Magic Sets errors. Errors in Magic Sets typically signal
    that the magic sets algorithm cannot be applied to the current datalog
    program. Subclasses should specify the reason why the algorithm cannot be
    applied, for instance if negations are present in the code.
    """
    pass


class BoundAggregationApplicationError(InvalidMagicSetError):
    """
    Magic Sets algorithm does not work if an argument which is an aggregate
    function is bound.

    Example
    -------
    Q(x, count(y)) :- P(x, y)
    ans(x) :- Q(x, 3)
    """
    pass


class NegationInMagicSetsRewriteError(InvalidMagicSetError):
    """
    Magic Sets algorithm does not work if negations are present in the code.
    """
    pass


class NonConjunctiveAntecedentInMagicSetsError(InvalidMagicSetError):
    """
    Magic Sets algorithm does not work if one of the rules has a non
    conjunctive antecedent.
    """
    pass


class NoConstantPredicateFoundError(InvalidMagicSetError):
    """
    Magic Sets algorithm only works if there is at least one predicate in the
    code with a constant as an argument.
    """
    pass


import operator
import typing

import pytest

from ...datalog.aggregation import (
    AggregationApplication,
    DatalogWithAggregationMixin,
    TranslateToLogicWithAggregation,
)
from ...datalog.basic_representation import DatalogProgram
from ...exceptions import ForbiddenExpressionError
from ...expression_walker import ExpressionWalker
from ...expressions import Constant, Symbol
from ...logic import Conjunction, Implication, Union
from ...type_system import Unknown
from ...utils.testing.logic import logic_exp_commutative_equal
from ..expression_processing import (
    CollapseConjunctiveAntecedents,
    RemoveDuplicatedAntecedentPredicates,
    TranslateToLogic,
    UnifyVariableEqualitiesMixin,
)

P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
Z = Symbol("Z")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")
a = Constant("a")
b = Constant("b")
c = Constant("c")

EQ = Constant(operator.eq)


class DatalogWithVariableEqualityUnification(
    TranslateToLogic,
    CollapseConjunctiveAntecedents,
    UnifyVariableEqualitiesMixin,
    RemoveDuplicatedAntecedentPredicates,
    DatalogProgram,
    ExpressionWalker,
):
    pass


def test_one_conjunct():
    rule = Implication(R(x), Conjunction((P(x, y), EQ(y, a))))
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    assert R in program.intensional_database()
    expected = Union((Implication(R(x), P(x, a)),))
    result = program.intensional_database()[R]
    assert result == expected


def test_two_conjuncts():
    rule = Implication(R(x, y), Conjunction((P(x, y), EQ(y, a), Q(y, y))))
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    expected = Union((Implication(R(x, a), Conjunction((P(x, a), Q(a, a)))),))
    assert R in program.intensional_database()
    result = program.intensional_database()[R]
    assert logic_exp_commutative_equal(result, expected)


def test_single_equality_antecedent():
    rule = Implication(R(x), Conjunction((EQ(x, a),)))
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    assert R not in program.intensional_database()
    assert R in program.extensional_database()


def test_between_vars_equalities():
    rule = Implication(R(x, y), Conjunction((P(x, y), EQ(y, x), Q(y, y))))
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    assert R in program.intensional_database()
    expecteds = [
        Union((Implication(R(x, x), Conjunction((P(x, x), Q(x, x)))),)),
        Union((Implication(R(y, y), Conjunction((P(y, y), Q(y, y)))),)),
    ]
    result = program.intensional_database()[R]
    assert any(
        logic_exp_commutative_equal(result, expected) for expected in expecteds
    )


def test_multiple_between_vars_equalities():
    rule = Implication(
        R(x, y, z), Conjunction((P(z, x), EQ(y, z), EQ(y, x), Q(y, y)))
    )
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    expecteds = [
        Union((Implication(R(v, v, v), Conjunction((P(v, v), Q(v, v)))),))
        for v in (x, y, z)
    ]
    assert R in program.intensional_database()
    result = program.intensional_database()[R]
    assert any(
        logic_exp_commutative_equal(result, expected) for expected in expecteds
    )


def test_mix_between_var_eqs_var_to_const_eq():
    rule = Implication(
        R(x, y, z),
        Conjunction((P(z, x), EQ(y, z), EQ(y, a), Q(y, y), EQ(b, x))),
    )
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    expected = Union(
        (
            Implication(
                R(b, a, a),
                Conjunction((P(a, b), Q(a, a))),
            ),
        )
    )
    assert R in program.intensional_database()
    result = program.intensional_database()[R]
    assert logic_exp_commutative_equal(result, expected)


def test_collapsable_conjunction():
    rule = Implication(
        R(x, y),
        Conjunction(
            (
                Conjunction((P(z), Q(y, x), EQ(y, z))),
                Conjunction((Q(z, y), P(z), P(y), EQ(z, a))),
                Z(b),
                Z(z),
            )
        ),
    )
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    expected = Union(
        (
            Implication(
                R(x, a),
                Conjunction((P(a), Q(a, x), Q(a, a), Z(b), Z(a))),
            ),
        )
    )
    result = program.intensional_database()[R]
    assert logic_exp_commutative_equal(result, expected)


def test_extra_var_eq_const_eq():
    rule = Implication(
        Z(x, y),
        Conjunction((Q(x, y), EQ(z, y), EQ(y, a))),
    )
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    expected = Union((Implication(Z(x, a), Q(x, a)),))
    result = program.intensional_database()[Z]
    assert logic_exp_commutative_equal(result, expected)
    rule = Implication(
        Z(x, y, z),
        Conjunction((Q(x, y), EQ(z, y), EQ(y, a))),
    )
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    expected = Union((Implication(Z(x, a, a), Q(x, a)),))
    result = program.intensional_database()[Z]
    assert logic_exp_commutative_equal(result, expected)


class DatalogWithVariableEqualityUnificationAndAggregation(
    TranslateToLogicWithAggregation,
    UnifyVariableEqualitiesMixin,
    RemoveDuplicatedAntecedentPredicates,
    DatalogWithAggregationMixin,
    DatalogProgram,
    ExpressionWalker,
):
    def function_sum(
        self,
        x: typing.AbstractSet,
        y: typing.AbstractSet,
    ) -> Unknown:
        return sum(v + w for v, w in zip(x, y))


def test_aggregation_symb_to_symb_eq():
    rule = Implication(
        P(x, Constant(sum)(y)),
        Conjunction(
            (
                R(x, y, z),
                EQ(y, z),
            )
        ),
    )
    program = DatalogWithVariableEqualityUnificationAndAggregation()
    program.walk(rule)
    expected = Union(
        (
            Implication(
                P(x, AggregationApplication(Constant(sum), (z,))), R(x, z, z)
            ),
        )
    )
    result = program.intensional_database()[P]
    assert logic_exp_commutative_equal(result, expected)


def test_aggregation_symb_to_const_eq():
    rule = Implication(
        P(x, Constant(sum)(y)),
        Conjunction(
            (
                R(x, y, z),
                EQ(y, a),
            )
        ),
    )
    program = DatalogWithVariableEqualityUnificationAndAggregation()
    with pytest.raises(ForbiddenExpressionError):
        program.walk(rule)


def test_unify_vareq_builtin():
    some_builtin = Constant[typing.Callable[..., str]](
        lambda s: s + s,
        verify_type=False,
        auto_infer_type=False,
    )
    rule = Implication(
        P(x, y),
        Conjunction((Q(x), Q(y), EQ(x, y), EQ(y, some_builtin(x)))),
    )
    program = DatalogWithVariableEqualityUnification()
    program.walk(rule)
    expected = Union(
        (Implication(P(y, y), Conjunction((Q(y), EQ(y, some_builtin(y))))),)
    )
    result = program.intensional_database()[P]
    assert logic_exp_commutative_equal(result, expected)



from ..expression_walker import (
    ExpressionWalker,
    PatternWalker,
    ResolveSymbolMixin,
    add_match
)
from ..expressions import Constant, Symbol
from ..relational_algebra import (
    ColumnStr,
    ExtendedProjection,
    FunctionApplicationListMember,
    NameColumns,
    NumberColumns,
    Projection,
    RelationalAlgebraSolver,
    int2columnint_constant,
    str2columnstr_constant
)
from ..relational_algebra_provenance import (
    BuildProvenanceAlgebraSetWalkIntoMixin,
    ProvenanceAlgebraSet,
    RelationalAlgebraProvenanceCountingSolverMixin
)
from .probabilistic_ra_utils import (
    DeterministicFactSet,
    ProbabilisticChoiceSet,
    ProbabilisticFactSet
)


class ProbSemiringSolverMixin(
    BuildProvenanceAlgebraSetWalkIntoMixin,
    RelationalAlgebraProvenanceCountingSolverMixin,
    PatternWalker
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.translated_probfact_sets = dict()

    @add_match(
        Projection,
        lambda exp: (
            isinstance(
                exp.relation,
                (
                    ProbabilisticFactSet,
                    ProbabilisticChoiceSet,
                ),
            )
        ) and (
            exp.relation.probability_column == int2columnint_constant(0)
        )
    )
    def eliminate_superfluous_projection(self, expression):
        new_relation = expression.relation.apply(
            Projection(
                expression.relation.relation,
                (expression.relation.probability_column,) +
                tuple(
                    int2columnint_constant(a.value + 1)
                    for a in expression.attributes
                )
            ),
            expression.relation.probability_column
        )
        return self.walk(new_relation)

    @add_match(Projection(DeterministicFactSet, ...))
    def push_projection_in_deterministic(self, expression):
        return self.walk(
            DeterministicFactSet(
                Projection(
                    expression.relation.relation,
                    expression.attributes
                ))
        )

    @add_match(
        DeterministicFactSet(Constant),
        lambda e: e.relation.value.is_empty()
    )
    def empty_deterministic_fact_set(self, deterministic_set):
        provenance_column = ColumnStr(Symbol.fresh().name)
        return ProvenanceAlgebraSet(
            deterministic_set.relation,
            str2columnstr_constant(provenance_column)
        )

    @add_match(DeterministicFactSet(Symbol))
    def deterministic_fact_set(self, deterministic_set):
        relation_symbol = deterministic_set.relation
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        relation = self.walk(relation_symbol)
        if isinstance(relation, Constant):
            columns = relation.value.columns
        else:
            columns = relation.columns()
        named_columns = tuple(
            str2columnstr_constant(f"col_{i}") for i in range(len(columns))
        )
        projection_list = [
            FunctionApplicationListMember(c, c)
            for c in named_columns
        ]

        prov_column = str2columnstr_constant(Symbol.fresh().name)
        provenance_set = NumberColumns(
            ExtendedProjection(
                NameColumns(relation, named_columns),
                tuple(projection_list)
                + (
                    FunctionApplicationListMember(
                        Constant[float](1.0),
                        prov_column,
                    ),
                ),
            ),
            (prov_column,) + named_columns
        )

        self.translated_probfact_sets[relation_symbol] = \
            ProvenanceAlgebraSet(
                provenance_set, int2columnint_constant(0)
            )
        return self.translated_probfact_sets[relation_symbol]

    @add_match(ProbabilisticFactSet(Symbol, ...))
    def probabilistic_fact_set(self, prob_fact_set):
        relation_symbol = prob_fact_set.relation
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        relation = self.walk(relation_symbol)
        self.translated_probfact_sets[relation_symbol] = \
            ProvenanceAlgebraSet(
                relation, prob_fact_set.probability_column
            )
        return self.translated_probfact_sets[relation_symbol]

    @add_match(ProbabilisticChoiceSet(Symbol, ...))
    def probabilistic_choice_set(self, prob_choice_set):
        return self.probabilistic_fact_set(prob_choice_set)

    @add_match(DeterministicFactSet)
    def deterministic_fact_set_general(self, det_fact_set):
        new_symbol = Symbol.fresh()
        self.symbol_table[new_symbol] = det_fact_set.relation
        return self.walk(DeterministicFactSet(new_symbol))

    @add_match(ProbabilisticFactSet)
    def probabilistic_fact_set_invalid(self, prob_fact_set):
        new_symbol = Symbol.fresh()
        self.symbol_table[new_symbol] = prob_fact_set.relation
        return self.walk(ProbabilisticFactSet(
            new_symbol, prob_fact_set.probability_column
        ))

    @add_match(ProbabilisticChoiceSet)
    def probabilistic_choice_set_to_symbol(self, prob_choice_set):
        new_symbol = Symbol.fresh()
        self.symbol_table[new_symbol] = prob_choice_set.relation
        return self.walk(ProbabilisticChoiceSet(
            new_symbol, prob_choice_set.probability_column
        ))

    @staticmethod
    def _check_prov_col_not_in_proj_list(provset, proj_list):
        if any(
            member.dst_column.value == provset.provenance_column
            for member in proj_list
        ):
            raise ValueError(
                "Cannot project on provenance column: "
                f"{provset.provenance_column}"
            )

    @staticmethod
    def _check_all_non_prov_cols_in_proj_list(provset, proj_list):
        if provset.value.is_empty():
            return
        non_prov_cols = set(provset.non_provenance_columns)
        found_cols = set(
            member.dst_column.value
            for member in proj_list
            if member.dst_column.value in non_prov_cols
            and member.fun_exp == member.dst_column
        )
        if non_prov_cols.symmetric_difference(found_cols):
            raise ValueError(
                "All non-provenance columns must be part of the extended "
                "projection as {c: c} projection list member."
            )


class ProbSemiringSolver(
    ProbSemiringSolverMixin,
    RelationalAlgebraSolver
):
    pass


class ProbSemiringToRelationalAlgebraSolver(
    ProbSemiringSolverMixin,
    ResolveSymbolMixin,
    ExpressionWalker
):
    def __init__(self, *args, symbol_table=None, **kwargs):
        self.symbol_table = symbol_table
        super().__init__(*args, **kwargs)



import json
import urllib.parse
from concurrent.futures import Future
from typing import AbstractSet, Tuple
from unittest.mock import MagicMock, create_autospec
from uuid import uuid4

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pytest
import tornado.testing
import tornado.websocket
from neurolang.exceptions import NeuroLangException
from neurolang.type_system import Unknown
from neurolang.utils.relational_algebra_set import (
    NamedRelationalAlgebraFrozenSet,
)

from ..app import Application
from ..queries import NeurolangQueryManager


@pytest.fixture
def future():
    future = create_autospec(Future)
    future.cancelled.return_value = False
    future.done.return_value = False
    future.running.return_value = True
    return future


@pytest.fixture
def error():
    return NeuroLangException("Something went wrong")


@pytest.fixture
def data():
    data = [
        (5, "dog", frozenset({(1, 2), (5, 6)})),
        (10, "cat", frozenset({(5, 6), (8, 9)})),
        (-5.25, "mouse", frozenset({(8, 9), (12, 13)})),
    ]
    return data


@pytest.fixture
def result(data):
    ans = NamedRelationalAlgebraFrozenSet(("a", "b", "c"), data)
    ans.row_type = Tuple[float, str, AbstractSet[Unknown]]
    results = {"ans": ans}
    return results


@pytest.fixture
def results(data):
    ans = NamedRelationalAlgebraFrozenSet(("a", "b", "c"), data)
    ans.row_type = Tuple[float, str, AbstractSet[Unknown]]
    voxels = NamedRelationalAlgebraFrozenSet(
        ("i", "j", "k"), [[12, 15, 98], [107, 2, 33], [89, 8, 34]]
    )
    voxels.row_type = Tuple[int, int, int]
    results = {"ans": ans, "Voxel": voxels}
    return results


@pytest.fixture
def figures(data):
    ans = NamedRelationalAlgebraFrozenSet(("a", "b", "c"), data)
    ans.row_type = Tuple[float, str, AbstractSet[Unknown]]
    N_points = 10000
    n_bins = 20
    x = np.random.randn(N_points)
    y = 0.4 * x + np.random.randn(N_points) + 5
    figx, ax = plt.subplots()
    ax.hist(x, bins=n_bins)
    plt.close(figx)
    figy, axy = plt.subplots()
    axy.hist(y, bins=n_bins)
    plt.close(figy)

    hists = NamedRelationalAlgebraFrozenSet(("agg_hist",), [[figx], [figy]])
    hists.row_type = Tuple[matplotlib.figure.Figure]
    figures = {"Hists": hists}
    return figures


@pytest.fixture
def nqm():
    mock_nqm = MagicMock(spec=NeurolangQueryManager)
    return mock_nqm


class TornadoTestCase(tornado.testing.AsyncHTTPTestCase):
    """
    unittest.TestCase inheriting from tornado's testing.AsyncHTTPTestCase
    to create the app under test and benefit from the convenience methods
    to generate requests for it.

    It's a bit hackish but works for writing pytest tests using the
    tornado.testing utilities.
    """

    def __init__(self, nqm) -> None:
        super().__init__()
        self.nqm = nqm

    def get_app(self):
        return Application(self.nqm)

    def runTest(self):
        pass


@pytest.fixture
def test_case(nqm):
    test_case = TornadoTestCase(nqm)
    test_case.setUp()

    yield test_case

    test_case.tearDown()


def test_status_handler_calls_get_result_on_nqm(
    test_case, nqm, future, results
):
    uuid = uuid4()
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = results
    nqm.get_result.return_value = future

    response = test_case.fetch(f"/v1/status/{uuid}")
    assert response.code == 200
    nqm.get_result.assert_called_with(str(uuid))


def test_status_handler_passes_params_to_query_results(
    test_case, nqm, future, results
):
    uuid = uuid4()
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = results
    nqm.get_result.return_value = future

    params = {"symbol": "ans", "start": 1, "length": 1, "sort": 1, "asc": 0}

    response = test_case.fetch(
        f"/v1/status/{uuid}?" + urllib.parse.urlencode(params)
    )
    assert response.code == 200
    qr = json.loads(response.body)
    assert qr["status"] == "ok"
    data = qr["data"]
    assert data["start"] == params["start"]
    assert data["length"] == params["length"]
    assert data["sort"] == params["sort"]
    assert data["asc"] == bool(params["asc"])
    assert data["uuid"] == str(uuid)
    assert data["done"] == True
    assert len(data["results"]) == 1
    ans = data["results"]["ans"]
    assert ans["columns"] == list(results["ans"].columns)
    assert ans["size"] == 3
    assert len(ans["values"]) == 1


def test_symbols_handler_calls_get_symbols_on_nqm(
    test_case, nqm, future, results
):
    engine = "neuroquery"
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = results
    nqm.get_symbols.return_value = future

    response = test_case.fetch(f"/v1/symbol/{engine}")
    assert response.code == 200
    nqm.get_symbols.assert_called_with(engine)


def test_figure_handler_gets_figure(test_case, nqm, future, figures):
    uuid = uuid4()
    future.done.return_value = True
    future.exception.return_value = None
    future.result.return_value = figures
    nqm.get_result.return_value = future

    params = {"symbol": "Hists", "row": 1, "col": 0}

    response = test_case.fetch(
        f"/v1/figure/{uuid}?" + urllib.parse.urlencode(params)
    )
    assert response.code == 200
    nqm.get_result.assert_called_with(str(uuid))



from functools import lru_cache
from itertools import product

import numpy as np
from scipy.linalg import kron

from .config import config
from .interval_algebra import (v_before, v_during, v_equals, v_finishes,
                               v_meets, v_overlaps, v_starts)
from .regions import ExplicitVBR, ImplicitVBR, Region

__all__ = ['cardinal_relation']

directions_dim_space = {
    'L': [0],
    'R': [0],
    'P': [1],
    'A': [1],
    'O': [0, 1, 2],
    'I': [2],
    'S': [2],
    'F': [3]
}

matrix_positions_per_directions = {
    'L': [0],
    'O': [1],
    'R': [2],
    'P': [0],
    'A': [2],
    'I': [0],
    'S': [2],
    'F': [2]
}

inverse_directions = {
    'R': 'L',
    'A': 'P',
    'S': 'I',
    'O': 'O',
    'L': 'R',
    'P': 'A',
    'I': 'S'
}

anatomical_restric_axes = directions_dim_space["A"] + directions_dim_space["S"]
anatomical_restrict_dirs = ["A", "P", "I", "S"]
anatomical_restrict_threshold = config["DEFAULT"].getfloat(
    "cardinalRelationThreshold"
)

relations = [
    v_before, v_overlaps, v_during, v_meets, v_starts, v_finishes, v_equals
]


def cardinal_relation(
    region,
    reference_region,
    directions,
    refine_overlapping=False,
    stop_at=None
):
    if region is reference_region:
        return False

    if type(region) is Region and type(reference_region) is Region:
        mat = direction_matrix(region.bounding_box,
                               reference_region.bounding_box)
        return is_in_direction(mat, directions)

    region, reference_region = cardinal_relation_prepare_regions(
        region, reference_region
    )

    if region == reference_region:
        result = directions == 'O'
    else:
        mat = direction_matrix(
            region.bounding_box, reference_region.bounding_box
        )
        dir_counts = None
        if not (refine_overlapping and "O" in directions) and is_in_direction(
            mat, directions
        ):
            result = True
        else:
            overlap = is_in_direction(mat, 'O')
            if (
                    overlap and
                    refine_overlapping and
                    isinstance(region, ExplicitVBR)
            ):
                mat, dir_counts = overlap_resolution(
                    region, reference_region, directions, stop_at
                )
            result = is_in_direction(mat, directions, dir_counts)

    return result


def cardinal_relation_prepare_regions(region, reference_region):
    if isinstance(region, ImplicitVBR):
        if isinstance(reference_region, ImplicitVBR):
            raise NotImplementedError(
                f'Comparison between two implicit regions '
                f'can\'t be performed: {region}, {reference_region}'
            )
        region = region.to_explicit_vbr(
            reference_region.affine, reference_region.image_dim
        )
    if isinstance(reference_region, ImplicitVBR):
        reference_region = reference_region.to_explicit_vbr(
            region.affine, region.image_dim
        )

    return region, reference_region


def overlap_resolution(
    region, reference_region, directions=None, stop_at=None
):
    if stop_at == 0:
        raise ValueError("stop_at must be larger than 0")

    region_stack = [
        (region.aabb_tree.root, reference_region.aabb_tree.root, 0)
    ]

    directions = directions.replace('O', '')
    if len(directions) == 0:
        directions = None

    total_mat = direction_matrix(
        region_stack[0][0].box, region_stack[0][1].box
    ) * 0
    dir_counts = np.zeros(len(anatomical_restrict_dirs), dtype=int)
    while region_stack:
        region, reference_region, level = region_stack.pop()

        mat = direction_matrix(region.box, reference_region.box, True)
        total_mat += mat
        dir_counts += np.array(
            [is_in_direction(mat, d) for d in anatomical_restrict_dirs], dtype=int
        )

        if (
            is_in_direction(mat, 'O') and
            (stop_at is None or level < stop_at - 1)
        ):
            region_nl = region.children
            ref_region_nl = reference_region.children
            if len(region_nl) == 0 and len(ref_region_nl) > 0:
                region_nl = (region,)
            elif len(region_nl) > 0 and len(ref_region_nl) == 0:
                ref_region_nl = (reference_region,)

            stack_update = [
                (r, ref, level + 1)
                for r, ref in product(region_nl, ref_region_nl)
            ]

            if len(stack_update) > 0:
                total_mat -= mat
                region_stack += stack_update
            elif directions is None:
                break

    return total_mat.clip(0, 1), dir_counts


@lru_cache(maxsize=128)
def is_in_direction_indices(n, direction):
    indices = [[0, 1, 2]] * n
    for i in direction:
        for dim in directions_dim_space[i]:
            indices[n - 1 - dim] = matrix_positions_per_directions[i]
    return np.ix_(*indices)


def is_in_direction(matrix, direction, dir_counts=None):
    """
    Returns whether the cardinal relation matrix describing the relative
    position of two regions is in the given directions.

    If a direction_counts array is provided, this method checks whether the
    amount of time a direction is present compared to its opposite is greater
    than a specific threshold.

    Parameters
    ----------
    matrix : numpy.array
        the cardinal relation matrix provided by the `direction_matrix`
        function
    direction : str
        a string containing the directions queried
    dir_counts : numpy.array, optional
        an array containing the counts of regions in a direction X for each
        direction of the anatomical_restrict_dirs

    Returns
    -------
    bool
        whether the regions described by the matrix are in the query directions
    """
    if dir_counts is not None:
        for d in direction:
            if d in anatomical_restrict_dirs:
                # For direction on the A/P or I/S axis, we consider a region is
                # in direction A/P if there are more than 80% subregions in
                # direction A/P.
                c0, c1 = (
                    dir_counts[anatomical_restrict_dirs.index(d)],
                    dir_counts[
                        anatomical_restrict_dirs.index(inverse_directions[d])
                    ],
                )
                ratio = c0 / (c0 + c1)
                if ratio > anatomical_restrict_threshold:
                    return True
                else:
                    direction = direction.replace(d, "")

        if len(direction) == 0:
            return False

    indices = is_in_direction_indices(matrix.ndim, direction)
    return np.any(matrix[indices] == 1)


def relation_vectors(intervals, other_region_intervals):
    obtained_vectors = np.empty(
        (min(len(intervals), len(other_region_intervals)), 3)
    )
    ov = 0
    for interval, other_region_interval in zip(
        intervals, other_region_intervals
    ):
        for f in relations:
            vector = f(interval, other_region_interval)
            if vector is not None:
                obtained_vectors[ov] = vector
                ov += 1
                break
    return obtained_vectors[:ov]


def direction_matrix(
    region_bb, another_region_bb, restrict_axes_directions=False
):
    rp_vector = relation_vectors(region_bb.limits, another_region_bb.limits)
    if restrict_axes_directions:
        restrict_axis_relation_to_overlaping_regions(rp_vector)
    tensor = rp_vector[0].reshape(1, 3)
    for i in range(1, len(rp_vector)):
        tensor = kron(
            rp_vector[i].reshape((3, ) + (1, ) * i),
            tensor
        ).squeeze()
    return tensor.clip(0, 1)


def restrict_axis_relation_to_overlaping_regions(rp_vector):
    """
    When checking for directions on the anterior/posterior or on the
    inferior/posterior axis, we only consider intervals which overlap on the
    other axis.
    """
    # for each axis
    for d, orth_dir in zip(
        anatomical_restric_axes, anatomical_restric_axes[::-1]
    ):
        # check that the other axis is overlaping
        if not rp_vector[orth_dir, 1]:
            # otherwise remove non-overlaping relations from current axis
            rp_vector[d] = (0, 1, 0)



from ..exceptions import (
    NeuroLangException,
    UnsupportedQueryError,
    UnsupportedSolverError,
)


class DistributionDoesNotSumToOneError(NeuroLangException):
    pass


class MalformedProbabilisticTupleError(NeuroLangException):
    pass


class NotHierarchicalQueryException(UnsupportedSolverError):
    pass


class UncomparableDistributionsError(NeuroLangException):
    pass


class NotEasilyShatterableError(UnsupportedSolverError):
    pass


class UnsupportedProbabilisticQueryError(UnsupportedQueryError):
    pass


class ForbiddenConditionalQueryNoProb(UnsupportedQueryError):
    pass


class ForbiddenConditionalQueryNonConjunctive(UnsupportedQueryError):
    pass


class RepeatedTuplesInProbabilisticRelationError(
    MalformedProbabilisticTupleError
):
    def __init__(self, n_repeated_tuples, n_tuples, message):
        self.n_repeated_tuples = n_repeated_tuples
        self.n_tuples = n_tuples
        self.message = message
        super().__init__(self.message)



import operator

import pytest

from .....datalog.basic_representation import DatalogProgram
from .....exceptions import UnsupportedProgramError
from .....expressions import Constant, Symbol
from .....logic import Conjunction, Implication
from ..spatial import (
    DetectEuclideanDistanceBoundMatrix,
    TranslateEuclideanDistanceBoundMatrixMixin,
)

EQ = Constant(operator.eq)
LT = Constant(operator.lt)
LE = Constant(operator.le)
GT = Constant(operator.gt)
GE = Constant(operator.ge)
P = Symbol("P")
EUCLIDEAN = Symbol("EUCLIDEAN")
i1, j1, k1 = Symbol("i1"), Symbol("j1"), Symbol("k1")
i2, j2, k2 = Symbol("i2"), Symbol("j2"), Symbol("k2")
FirstRange = Symbol("FirstRange")
SecondRange = Symbol("SecondRange")
d = Symbol("d")


def test_euclidean_spatial_bound_detection():
    max_dist = Constant(2)
    detector = DetectEuclideanDistanceBoundMatrix()
    expression = Implication(
        P(i2, j2, k2),
        Conjunction(
            (
                FirstRange(i1, j1, k1),
                SecondRange(i2, j2, k2),
                EQ(d, EUCLIDEAN(i1, j1, k1, i2, j2, k2)),
                LT(d, max_dist),
            )
        ),
    )
    assert detector.walk(expression)
    expression = Implication(
        P(i2, j2, k2),
        Conjunction(
            (
                FirstRange(i1, j1, k1),
                SecondRange(i2, j2, k2),
                EQ(d, EUCLIDEAN(i1, j1, k1, i2, j2, k2)),
                LE(d, max_dist),
            )
        ),
    )
    assert detector.walk(expression)
    expression = Implication(
        P(i2, j2, k2),
        Conjunction(
            (
                FirstRange(i1, j1, k1),
                SecondRange(i2, j2, k2),
                EQ(d, EUCLIDEAN(i1, j1, k1, i2, j2, k2)),
                GE(max_dist, d),
            )
        ),
    )
    assert detector.walk(expression)


def test_reversed_equality():
    max_dist = Constant(10)
    detector = DetectEuclideanDistanceBoundMatrix()
    expression = Implication(
        P(i2, j2, k2),
        Conjunction(
            (
                FirstRange(i1, j1, k1),
                SecondRange(i2, j2, k2),
                EQ(EUCLIDEAN(i1, j1, k1, i2, j2, k2), d),
                LT(d, max_dist),
            )
        ),
    )
    assert detector.walk(expression)
    expression = Implication(
        P(i2, j2, k2),
        Conjunction(
            (
                FirstRange(i1, j1, k1),
                SecondRange(i2, j2, k2),
                EQ(EUCLIDEAN(i1, j1, k1, i2, j2, k2), d),
                GT(max_dist, d),
            )
        ),
    )
    assert detector.walk(expression)


def test_unsupported_unbounded_distance():
    detector = DetectEuclideanDistanceBoundMatrix()
    expression = Implication(
        P(i2, j2, k2),
        Conjunction(
            (
                FirstRange(i1, j1, k1),
                SecondRange(i2, j2, k2),
                EQ(EUCLIDEAN(i1, j1, k1, i2, j2, k2), d),
            )
        ),
    )
    with pytest.raises(UnsupportedProgramError):
        detector.walk(expression)


class ProgramTest(TranslateEuclideanDistanceBoundMatrixMixin, DatalogProgram):
    pass


def test_unsupported_non_extensional_safe_range_pred():
    max_dist = Constant(2)
    translator = ProgramTest()
    expression = Implication(
        P(i2, j2, k2),
        Conjunction(
            (
                FirstRange(i1, j1, k1),
                SecondRange(i2, j2, k2),
                EQ(EUCLIDEAN(i1, j1, k1, i2, j2, k2), d),
                GT(max_dist, d),
            )
        ),
    )
    with pytest.raises(NotImplementedError):
        translator.walk(expression)



from ...expressions import Symbol
from ...logic import (
    Conjunction,
    Disjunction,
    ExistentialPredicate
)
from ..containment import is_contained


def test_containment_conjunctive_query():
    R = Symbol('R')
    S = Symbol('S')
    w = Symbol('w')
    x = Symbol('x')
    y = Symbol('y')

    r1 = ExistentialPredicate(y, ExistentialPredicate(
        w,
        Conjunction((R(x, y), R(x, w)))
    ))

    r2 = ExistentialPredicate(y, R(x, y))

    r3 = ExistentialPredicate(y, S(x, y))

    r4 = ExistentialPredicate(y, ExistentialPredicate(
        w,
        Conjunction((R(x, y), S(x, w)))
    ))

    assert is_contained(r1, r1)
    assert is_contained(r2, r1)
    assert is_contained(r1, r2)
    assert not is_contained(r1, r3)
    assert not is_contained(r3, r1)
    assert is_contained(r1, r4)
    assert is_contained(r3, r4)
    assert not is_contained(r4, r1)
    assert not is_contained(r4, r3)


def test_containment_disjunctive_query():
    R = Symbol('R')
    S = Symbol('S')
    w = Symbol('w')
    x = Symbol('x')
    y = Symbol('y')

    r1 = ExistentialPredicate(y, ExistentialPredicate(
        w,
        Disjunction((R(x, y), R(x, w)))
    ))

    r2 = ExistentialPredicate(y, R(x, y))

    r3 = ExistentialPredicate(y, S(x, y))

    r4 = ExistentialPredicate(y, ExistentialPredicate(
        w,
        Disjunction((R(x, y), S(x, w)))
    ))

    assert is_contained(r1, r1)
    assert is_contained(r2, r1)
    assert is_contained(r1, r2)
    assert not is_contained(r1, r3)
    assert not is_contained(r3, r1)
    assert is_contained(r4, r1)
    assert is_contained(r4, r3)
    assert not is_contained(r1, r4)
    assert not is_contained(r3, r4)


def test_containment_query_cross_product():
    R = Symbol('R')
    x = Symbol('x')
    y = Symbol('y')

    r1 = R(x)
    r2 = R(y)

    assert not is_contained(r2, r1)
    assert not is_contained(r1, r2)


def test_containment_query():
    R = Symbol('R')
    x = Symbol('x')
    y = Symbol('y')

    r1 = ExistentialPredicate(
        y,
        Conjunction((R(x, y), R(y, x), R(x, x)))
    )
    r2 = ExistentialPredicate(y, Conjunction((R(x, y), R(y, x))))

    assert is_contained(r2, r1)
    assert not is_contained(r1, r2)


def test_containment_mixed():
    R = Symbol('R')
    S = Symbol('S')
    w = Symbol('w')
    x = Symbol('x')
    y = Symbol('y')
    z = Symbol('z')

    r1 = ExistentialPredicate(y, ExistentialPredicate(
        w,
        Conjunction((R(x, y), R(y, x), R(w, w)))
    ))

    r2 = ExistentialPredicate(
        y,
        Conjunction((R(x, y), R(y, x)))
    )

    r3 = ExistentialPredicate(
        y,
        R(x, y)
    )

    r4 = ExistentialPredicate(
        y,
        Disjunction((R(x, y), S(x, y)))
    )

    r5 = ExistentialPredicate(
        y,
        S(x, y)
    )

    r6 = ExistentialPredicate(y, ExistentialPredicate(
        z,
        Conjunction((Disjunction((R(x, y), S(x, y))), S(x, z)))
    ))

    assert is_contained(r1, r1)
    assert is_contained(r2, r1)
    assert not is_contained(r1, r2)
    assert is_contained(r3, r1)
    assert is_contained(r3, r1)
    assert is_contained(r4, r3)
    assert is_contained(r5, r6)
    assert not is_contained(r3, r5)
    assert not is_contained(r5, r3)
    assert is_contained(r6, r5)



from pathlib import Path
from typing import AbstractSet, Tuple
from unittest.mock import patch

import pandas as pd
import pytest

from .. import NeurolangDL
from ..neurosynth_utils import StudyID


@pytest.fixture
def metadata():
    return pd.DataFrame(
        [
            (
                9065511,
                "MNI",
            ),
            (
                9084599,
                "MNI",
            ),
            (
                9114263,
                "MNI",
            ),
            (
                9185551,
                "TAL",
            ),
            (
                9256495,
                "TAL",
            ),
        ],
        columns=["id", "space"],
    )


@pytest.fixture
def features():
    df = pd.DataFrame(
        [
            (9065511, 0.0, 0.0, 0.0, 0.079103, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
            (9084599, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
            (9114263, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
            (
                9185551,
                0.0,
                0.079103,
                0.0,
                0.0,
                0.0,
                0.0,
                0.124442,
                0.0,
                0.0,
                0.0,
            ),
            (9256495, 0.0, 0.0, 0.0, 0.0, 0.219273, 0.0, 0.0, 0.0, 0.0, 0.0),
        ],
        columns=[
            "id",
            "scanner",
            "periodicals",
            "composed",
            "adhd",
            "tms",
            "space",
            "stimulus response",
            "hemisphere",
            "neurofunctional",
            "youth",
        ],
    )
    df = df.set_index("id")
    return df


@pytest.fixture
def peak_data():
    df = pd.DataFrame(
        [
            (9114263, 27003, "1.", 517253, -38.0, -38.0, 69.0, "MNI"),
            (9256495, 40779, "1", 771542, 55.0, 28.0, 9.0, "TAL"),
            (9114263, 27003, "1.", 517461, -11.0, -20.0, 62.0, "MNI"),
            (9114263, 27003, "1.", 517255, -29.0, -29.0, 64.0, "MNI"),
            (9256495, 40779, "1", 771532, -14.0, -8.0, -18.0, "TAL"),
            (9114263, 27003, "1.", 517298, -18.0, -58.0, 62.0, "MNI"),
            (9114263, 27003, "1.", 517349, -28.0, -8.0, 64.0, "MNI"),
            (9114263, 27003, "1.", 517445, -63.0, -41.0, 38.0, "MNI"),
            (9114263, 27003, "1.", 517443, -63.0, -29.0, 31.0, "MNI"),
            (9114263, 27003, "1.", 517272, -5.0, -48.0, 52.0, "MNI"),
            (9114263, 27003, "1.", 517552, -2.0, -30.0, 36.0, "MNI"),
            (9084599, 27000, "4.", 517181, 12.0, 10.0, 16.0, "MNI"),
            (9256495, 40779, "1", 771581, 14.0, -56.0, -7.0, "TAL"),
            (9256495, 40779, "1", 771588, -20.0, -64.0, 9.0, "TAL"),
            (9114263, 27003, "1.", 517416, -39.0, -34.0, 18.0, "MNI"),
        ],
        columns=[
            "id",
            "table_id",
            "table_num",
            "peak_id",
            "x",
            "y",
            "z",
            "space",
        ],
    )
    return df


@pytest.fixture
def mni_peaks(peak_data):
    df = peak_data[["id", "x", "y", "z"]]
    return df.astype({"x": int, "y": int, "z": int})


@pytest.fixture
def term_data(features):
    term_data = pd.melt(
        features.reset_index(),
        var_name="term",
        id_vars="id",
        value_name="tfidf",
    )
    return term_data


@patch("neurolang.frontend.query_resolution.fetch_study_metadata")
def test_load_neurosynth_study_ids(mock_fetch_study_metadata, metadata):
    mock_fetch_study_metadata.return_value = metadata
    data_dir = Path("mock_data_dir")

    nl = NeurolangDL()
    symbol = nl.load_neurosynth_study_ids(data_dir, name="StudyIds")
    assert symbol.symbol_name == "StudyIds"
    assert set(nl[symbol.symbol_name].value) == set(
        (id,) for id in metadata["id"].values
    )

    symbol = nl.load_neurosynth_study_ids(data_dir, convert_study_ids=True)
    assert nl[symbol.symbol_name].type is AbstractSet[Tuple[StudyID]]


@patch("neurolang.frontend.query_resolution.get_ns_term_study_associations")
def test_load_neurosynth_term_study_associations(
    mock_get_term_data, term_data
):
    mock_get_term_data.return_value = term_data
    data_dir = Path("mock_data_dir")

    nl = NeurolangDL()
    symbol = nl.load_neurosynth_term_study_associations(
        data_dir, name="TermInStudyTFIDF"
    )
    assert symbol.symbol_name == "TermInStudyTFIDF"
    assert set(nl[symbol.symbol_name].value) == set(
        tuple(e) for e in term_data.values
    )

    symbol = nl.load_neurosynth_term_study_associations(
        data_dir, convert_study_ids=True
    )
    assert (
        nl[symbol.symbol_name].type is AbstractSet[Tuple[StudyID, str, float]]
    )


@patch("neurolang.frontend.query_resolution.get_ns_mni_peaks_reported")
def test_load_neurosynth_mni_peaks_reported(mock_peaks_reported, mni_peaks):
    mock_peaks_reported.return_value = mni_peaks
    data_dir = Path("mock_data_dir")

    nl = NeurolangDL()
    symbol = nl.load_neurosynth_mni_peaks_reported(
        data_dir, name="PeakReported"
    )
    assert symbol.symbol_name == "PeakReported"
    assert set(nl[symbol.symbol_name].value) == set(
        tuple(e) for e in mni_peaks.values
    )

    symbol = nl.load_neurosynth_mni_peaks_reported(
        data_dir, convert_study_ids=True
    )
    assert (
        nl[symbol.symbol_name].type
        is AbstractSet[Tuple[StudyID, int, int, int]]
    )



import pytest

from ....expressions import Constant, Symbol
from ....logic import Conjunction, Implication
from ..logic import logic_exp_commutative_equal

P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
x = Symbol("x")
y = Symbol("y")
a = Constant("a")


def test_compare_simple_expressions():
    assert logic_exp_commutative_equal(P, P)
    assert not logic_exp_commutative_equal(P, Q)
    with pytest.raises(ValueError):
        logic_exp_commutative_equal(P, 3)


def test_compare_conjunctions():
    conj1 = Conjunction((P(x, y), Q(x)))
    conj2 = Conjunction((Q(x), P(x, y)))
    assert logic_exp_commutative_equal(conj1, conj2)
    conj3 = Conjunction((Q(y), P(x, y)))
    assert not logic_exp_commutative_equal(conj1, conj3)
    conj4 = Conjunction((P(x, y), Q(y)))
    assert not logic_exp_commutative_equal(conj1, conj4)


def test_compare_implications():
    impl1 = Implication(R(x, a), Conjunction((Q(a, a), P(x, a))))
    impl2 = Implication(R(x, a), Conjunction((P(x, a), Q(a, a))))
    assert logic_exp_commutative_equal(impl1, impl2)



import pytest
from unittest.mock import MagicMock, ANY
from ... import expression_walker, expressions
from .. import DatalogProgram, Fact, Implication
from ..chase import ChaseNamedStratified, ChaseNR, ChaseSN
from ..chase.general import NoValidChaseClassForStratumException
from ..expressions import TranslateToLogic
from ..instance import MapInstance

C_ = expressions.Constant
S_ = expressions.Symbol
Imp_ = Implication
F_ = Fact
Eb_ = expressions.ExpressionBlock


class Datalog(
    TranslateToLogic,
    DatalogProgram,
    expression_walker.ExpressionBasicEvaluator,
):
    pass


x = S_("X")
y = S_("Y")
z = S_("Z")
anc = S_("anc")
par = S_("par")
q = S_("q")
a = C_("a")
b = C_("b")
c = C_("c")
d = C_("d")


@pytest.fixture
def datalog():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
            Imp_(q(x), anc(a, x)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)
    return dl


def test_stratification_resolution_works(datalog):
    solution = ChaseNamedStratified(datalog).build_chase_solution()
    assert solution[q].value == {C_((e,)) for e in (b, c, d)}


def test_stratified_chases(datalog):
    # Recursive stratum can be solved by just ChaseSN
    classes = (ChaseSN,)
    solution = ChaseNamedStratified(datalog, chase_classes=classes).build_chase_solution()
    assert solution[q].value == {C_((e,)) for e in (b, c, d)}

    # Recursive stratum cannot be solved by just ChaseNR
    classes = (ChaseNR,)
    with pytest.raises(
        NoValidChaseClassForStratumException,
    ):
        solution = ChaseNamedStratified(datalog, chase_classes=classes).build_chase_solution()

    # Check that each stratum is solved by the first possible class
    strata = list(datalog.intensional_database().values())
    spied_chase_sn = MagicMock(wraps=ChaseSN(datalog, strata[0]))
    spied_chase_nr = MagicMock(wraps=ChaseNR(datalog, strata[1]))
    classes = (
        MagicMock(return_value=spied_chase_nr),
        MagicMock(return_value=spied_chase_sn),
    )

    solution = ChaseNamedStratified(datalog, chase_classes=classes).build_chase_solution()
    assert solution[q].value == {C_((e,)) for e in (b, c, d)}
    spied_chase_sn.execute_chase.assert_called_with(
        list(strata[0].formulas),
        MapInstance(datalog.extensional_database()),
        ANY,
    )
    spied_chase_nr.execute_chase.assert_called_once()


def test_unstratifyable_code_resolution():
    Q = S_("Q")  # noqa: N806
    R = S_("R")  # noqa: N806
    S = S_("S")  # noqa: N806
    T = S_("T")  # noqa: N806
    x = S_("x")
    y = S_("y")
    code = Eb_(
        [
            F_(Q(C_(1), C_(2))),
            F_(Q(C_(2), C_(3))),
            Imp_(R(x, y), Q(x, y)),
            Imp_(R(x, y), Q(y, x)),
            Imp_(S(x), R(x, y) & T(y)),
            Imp_(T(x), R(x, y) & S(x)),
        ]
    )
    dl = Datalog()
    dl.walk(code)
    solution = ChaseNamedStratified(dl, chase_classes=(ChaseSN,)).build_chase_solution()
    assert solution[R].value == {(1, 2), (2, 1), (3, 2), (2, 3)}
